from .config import ConfigMapping
from .decorators import (
    composite_solid,
    daily_schedule,
    hourly_schedule,
    lambda_solid,
    monthly_schedule,
    pipeline,
    schedule,
    schedules,
    solid,
    weekly_schedule,
)
from .dependency import (
    DependencyDefinition,
    MultiDependencyDefinition,
    Solid,
    SolidHandle,
    SolidInputHandle,
    SolidInvocation,
    SolidOutputHandle,
)
from .environment_schema import (
    EnvironmentSchema,
    create_environment_schema,
    create_environment_type,
)
from .events import (
    EventMetadataEntry,
    ExpectationResult,
    Failure,
    JsonMetadataEntryData,
    MarkdownMetadataEntryData,
    Materialization,
    Output,
    PathMetadataEntryData,
    RetryRequested,
    TextMetadataEntryData,
    TypeCheck,
    UrlMetadataEntryData,
)
from .executable import ExecutablePipeline
from .executor import (
    ExecutorDefinition,
    default_executors,
    executor,
    in_process_executor,
    multiprocess_executor,
)
from .handle import ExecutionTargetHandle, LoaderEntrypoint
from .input import InputDefinition, InputMapping
from .logger import LoggerDefinition, logger
from .mode import ModeDefinition
from .output import OutputDefinition, OutputMapping
from .partition import Partition, PartitionSetDefinition, repository_partitions
from .pipeline import PipelineDefinition
from .preset import PresetDefinition
from .repository import RepositoryDefinition
from .resource import ResourceDefinition, resource
from .schedule import ScheduleDefinition, ScheduleExecutionContext
from .solid import CompositeSolidDefinition, ISolidDefinition, SolidDefinition
from .solid_container import IContainSolids, create_execution_structure
from .system_storage import SystemStorageData, SystemStorageDefinition, system_storage



import os

from dagster import Bool, Int, check
from dagster.config import Field, Permissive
from dagster.config.field import resolve_to_config_type
from dagster.config.validate import validate_config
from dagster.core.errors import DagsterInvalidConfigError
from dagster.utils import merge_dicts
from dagster.utils.yaml_utils import load_yaml_from_globs

DAGSTER_CONFIG_YAML_FILENAME = "dagster.yaml"


def dagster_instance_config(base_dir, config_filename=DAGSTER_CONFIG_YAML_FILENAME, overrides=None):
    overrides = check.opt_dict_param(overrides, 'overrides')
    dagster_config_dict = merge_dicts(
        load_yaml_from_globs(os.path.join(base_dir, config_filename)), overrides
    )
    dagster_config_type = resolve_to_config_type(define_dagster_config_cls())
    dagster_config = validate_config(dagster_config_type, dagster_config_dict)
    if not dagster_config.success:
        raise DagsterInvalidConfigError(
            'Errors whilst loading dagster instance config at {}.'.format(config_filename),
            dagster_config.errors,
            dagster_config_dict,
        )
    return dagster_config.value


def config_field_for_configurable_class():
    return Field({'module': str, 'class': str, 'config': Field(Permissive())}, is_required=False)


def define_dagster_config_cls():
    return {
        'local_artifact_storage': config_field_for_configurable_class(),
        'compute_logs': config_field_for_configurable_class(),
        'run_storage': config_field_for_configurable_class(),
        'event_log_storage': config_field_for_configurable_class(),
        'schedule_storage': config_field_for_configurable_class(),
        'scheduler': config_field_for_configurable_class(),
        'run_launcher': config_field_for_configurable_class(),
        'dagit': Field(
            {
                'execution_manager': Field(
                    {
                        'disabled': Field(Bool, is_required=False),
                        'max_concurrent_runs': Field(Int, is_required=False),
                    },
                    is_required=False,
                ),
            },
            is_required=False,
        ),
        'telemetry': Field({'enabled': Field(Bool, default_value=True, is_required=False)}),
    }



from collections import namedtuple

from dagster import check
from dagster.serdes import whitelist_for_serdes

from .config_type import ConfigScalarKind, ConfigType, ConfigTypeKind
from .field import Field


def get_recursive_type_keys(config_type_snap, config_schema_snapshot):
    check.inst_param(config_type_snap, 'config_type_snap', ConfigTypeSnap)
    check.inst_param(config_schema_snapshot, 'config_schema_snapshot', ConfigSchemaSnapshot)
    result_keys = set()
    for type_key in config_type_snap.get_child_type_keys():
        result_keys.add(type_key)
        for recurse_key in get_recursive_type_keys(
            config_schema_snapshot.get_config_snap(type_key), config_schema_snapshot
        ):
            result_keys.add(recurse_key)
    return result_keys


@whitelist_for_serdes
class ConfigSchemaSnapshot(namedtuple('_ConfigSchemaSnapshot', 'all_config_snaps_by_key')):
    def __new__(cls, all_config_snaps_by_key):
        return super(ConfigSchemaSnapshot, cls).__new__(
            cls,
            all_config_snaps_by_key=check.dict_param(
                all_config_snaps_by_key,
                'all_config_snaps_by_key',
                key_type=str,
                value_type=ConfigTypeSnap,
            ),
        )

    @property
    def all_config_keys(self):
        return list(self.all_config_snaps_by_key.keys())

    def get_config_snap(self, key):
        check.str_param(key, 'key')
        return self.all_config_snaps_by_key[key]

    def has_config_snap(self, key):
        check.str_param(key, 'key')
        return key in self.all_config_snaps_by_key


@whitelist_for_serdes
class ConfigTypeSnap(
    namedtuple(
        '_ConfigTypeSnap',
        'kind key given_name description '
        'type_param_keys '  # only valid for closed generics (Set, Tuple, List, Optional)
        'enum_values '  # only valid for enums
        'fields '  # only valid for dicts and selectors
        'scalar_kind',  # only valid for scalars
    )
):
    # serdes log
    # * Adding scalar_kind
    def __new__(
        cls,
        kind,
        key,
        given_name,
        description,
        type_param_keys,
        enum_values,
        fields,
        scalar_kind=None,  # Old version of object will not have this property
    ):
        return super(ConfigTypeSnap, cls).__new__(
            cls,
            kind=check.inst_param(kind, 'kind', ConfigTypeKind),
            key=check.str_param(key, 'key'),
            given_name=check.opt_str_param(given_name, 'given_name'),
            type_param_keys=None
            if type_param_keys is None
            else check.list_param(type_param_keys, 'type_param_keys', of_type=str),
            enum_values=None
            if enum_values is None
            else check.list_param(enum_values, 'enum_values', of_type=ConfigEnumValueSnap),
            fields=None
            if fields is None
            else sorted(
                check.list_param(fields, 'field', of_type=ConfigFieldSnap), key=lambda ct: ct.name
            ),
            description=check.opt_str_param(description, 'description'),
            scalar_kind=check.opt_inst_param(scalar_kind, 'scalar_kind', ConfigScalarKind),
        )

    @property
    def inner_type_key(self):
        # valid for Noneable and Array
        check.invariant(self.kind == ConfigTypeKind.NONEABLE or self.kind == ConfigTypeKind.ARRAY)
        check.invariant(len(self.type_param_keys) == 1)
        return self.type_param_keys[0]

    @property
    def scalar_type_key(self):
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        return self.type_param_keys[0]

    @property
    def non_scalar_type_key(self):
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        return self.type_param_keys[1]

    def _get_field(self, name):
        check.str_param(name, 'name')
        check.invariant(ConfigTypeKind.has_fields(self.kind))

        for f in self.fields:
            if f.name == name:
                return f

        return None

    def get_field(self, name):
        field = self._get_field(name)
        if not field:
            check.failed('Field {name} not found'.format(name=name))
        return field

    def has_field(self, name):
        return bool(self._get_field(name))

    @property
    def field_names(self):
        return [fs.name for fs in self.fields]

    def get_child_type_keys(self):
        if ConfigTypeKind.is_closed_generic(self.kind):
            return self.type_param_keys
        elif ConfigTypeKind.has_fields(self.kind):
            return [field.type_key for field in self.fields]
        else:
            return []

    def has_enum_value(self, value):
        check.invariant(self.kind == ConfigTypeKind.ENUM)
        for enum_value in self.enum_values:
            if enum_value.value == value:
                return True
        return False


@whitelist_for_serdes
class ConfigEnumValueSnap(namedtuple('_ConfigEnumValueSnap', 'value description')):
    def __new__(cls, value, description):
        return super(ConfigEnumValueSnap, cls).__new__(
            cls,
            value=check.str_param(value, 'value'),
            description=check.opt_str_param(description, 'description'),
        )


@whitelist_for_serdes
class ConfigFieldSnap(
    namedtuple(
        '_ConfigFieldSnap',
        'name type_key is_required default_provided default_value_as_json_str description',
    )
):
    def __new__(
        cls, name, type_key, is_required, default_provided, default_value_as_json_str, description
    ):
        return super(ConfigFieldSnap, cls).__new__(
            cls,
            name=check.opt_str_param(name, 'name'),
            type_key=check.str_param(type_key, 'type_key'),
            is_required=check.bool_param(is_required, 'is_required'),
            default_provided=check.bool_param(default_provided, 'default_provided'),
            default_value_as_json_str=check.opt_str_param(
                default_value_as_json_str, 'default_value_as_json_str'
            ),
            description=check.opt_str_param(description, 'description'),
        )


def snap_from_field(name, field):
    check.str_param(name, 'name')
    check.inst_param(field, 'field', Field)
    return ConfigFieldSnap(
        name=name,
        type_key=field.config_type.key,
        is_required=field.is_required,
        default_provided=field.default_provided,
        default_value_as_json_str=field.default_value_as_json_str
        if field.default_provided
        else None,
        description=field.description,
    )


def snap_from_config_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    return ConfigTypeSnap(
        key=config_type.key,
        given_name=config_type.given_name,
        kind=config_type.kind,
        description=config_type.description,
        type_param_keys=[ct.key for ct in config_type.type_params] if config_type.type_params
        # jam scalar union types into type_param_keys
        else [config_type.scalar_type.key, config_type.non_scalar_type.key]
        if config_type.kind == ConfigTypeKind.SCALAR_UNION
        else None,
        enum_values=[
            ConfigEnumValueSnap(ev.config_value, ev.description) for ev in config_type.enum_values
        ]
        if config_type.kind == ConfigTypeKind.ENUM
        else None,
        fields=[snap_from_field(name, field) for name, field in config_type.fields.items()]
        if ConfigTypeKind.has_fields(config_type.kind)
        else None,
        scalar_kind=config_type.scalar_kind if config_type.kind == ConfigTypeKind.SCALAR else None,
    )



import re
import uuid

import pytest

from dagster import PipelineDefinition, RunConfig, execute_pipeline, solid

EPS = 0.001


def test_default_run_id():
    called = {}

    @solid
    def check_run_id(context):
        called['yes'] = True
        assert uuid.UUID(context.run_id)
        called['run_id'] = context.run_id

    pipeline = PipelineDefinition(solid_defs=[check_run_id])

    result = execute_pipeline(pipeline)
    assert result.run_id == called['run_id']
    assert called['yes']


def test_provided_run_id():
    called = {}

    @solid
    def check_run_id(context):
        called['yes'] = True
        assert context.run_id == 'given'

    pipeline = PipelineDefinition(solid_defs=[check_run_id])

    result = execute_pipeline(pipeline, run_config=RunConfig(run_id='given'))
    assert result.run_id == 'given'

    assert called['yes']


def test_select_mode():
    called = {}

    @solid
    def check_run_id(context):
        called['yes'] = True
        assert uuid.UUID(context.run_id)
        called['run_id'] = context.run_id

    pipeline = PipelineDefinition(solid_defs=[check_run_id])

    with pytest.warns(
        UserWarning,
        match=re.escape(
            'In 0.8.0, the use of `run_config` to set pipeline mode and tags will be deprecated'
        ),
    ):
        result = execute_pipeline(pipeline, run_config=RunConfig(mode='default'))
    assert result.run_id == called['run_id']
    assert called['yes']


def test_injected_tags():
    called = {}

    @solid
    def check_tags(context):
        assert context.get_tag('foo') == 'bar'
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='injected_run_id', solid_defs=[check_tags])
    with pytest.warns(
        UserWarning,
        match=re.escape(
            'In 0.8.0, the use of `run_config` to set pipeline mode and tags will be deprecated'
        ),
    ):
        result = execute_pipeline(pipeline_def, run_config=RunConfig(tags={'foo': 'bar'}))

    assert result.success
    assert called['yup']


def test_pipeline_tags():
    called = {}

    @solid
    def check_tags(context):
        assert context.get_tag('foo') == 'bar'
        called['yup'] = True

    pipeline_def_with_tags = PipelineDefinition(
        name='injected_run_id', solid_defs=[check_tags], tags={'foo': 'bar'}
    )
    result = execute_pipeline(pipeline_def_with_tags)
    assert result.success
    assert called['yup']

    called = {}
    pipeline_def_with_override_tags = PipelineDefinition(
        name='injected_run_id', solid_defs=[check_tags], tags={'foo': 'notbar'}
    )
    result = execute_pipeline(pipeline_def_with_override_tags, tags={'foo': 'bar'})
    assert result.success
    assert called['yup']


def test_overwrite_tags():
    called = {}

    @solid
    def check_tags(context):
        assert context.get_tag('foo') == 'baz'
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='injected_run_id', solid_defs=[check_tags])
    with pytest.warns(
        UserWarning,
        match=re.escape(
            'In 0.8.0, the use of `run_config` to set pipeline mode and tags will be deprecated'
        ),
    ):
        result = execute_pipeline(
            pipeline_def, run_config=RunConfig(tags={'foo': 'bar'}), tags={'foo': 'baz'}
        )

    assert result.success
    assert called['yup']



from alembic import op
from sqlalchemy.engine import reflection


def get_inspector():
    # pylint: disable=no-member
    bind = op.get_context().bind
    return reflection.Inspector.from_engine(bind)


def get_table_names():
    return get_inspector().get_table_names()


def has_table(table_name):
    return table_name in get_table_names()


def has_column(table_name, column_name):
    if not has_table(table_name):
        return False
    columns = [x.get('name') for x in get_inspector().get_columns(table_name)]
    return column_name in columns



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_basic_solids_config 1'] = {
    'execution': {
        'in_process': {
            'config': {
                'marker_to_close': '',
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        },
        'multiprocess': {
            'config': {
                'max_concurrent': 0,
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        }
    },
    'loggers': {
        'console': {
            'config': {
                'log_level': '',
                'name': ''
            }
        }
    },
    'resources': {
    },
    'solids': {
        'required_field_solid': {
            'config': {
                'required_int': 0
            }
        }
    },
    'storage': {
        'filesystem': {
            'config': {
                'base_dir': ''
            }
        },
        'in_memory': {
        }
    }
}

snapshots['test_two_modes 1'] = {
    'resources': {
        'value': {
            'config': {
                'mode_one_field': ''
            }
        }
    }
}

snapshots['test_two_modes 2'] = {
    'execution': {
        'in_process': {
            'config': {
                'marker_to_close': '',
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        },
        'multiprocess': {
            'config': {
                'max_concurrent': 0,
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        }
    },
    'loggers': {
        'console': {
            'config': {
                'log_level': '',
                'name': ''
            }
        }
    },
    'resources': {
        'value': {
            'config': {
                'mode_one_field': ''
            }
        }
    },
    'solids': {
    },
    'storage': {
        'filesystem': {
            'config': {
                'base_dir': ''
            }
        },
        'in_memory': {
        }
    }
}

snapshots['test_two_modes 3'] = {
    'resources': {
        'value': {
            'config': {
                'mode_two_field': 0
            }
        }
    }
}

snapshots['test_two_modes 4'] = {
    'execution': {
        'in_process': {
            'config': {
                'marker_to_close': '',
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        },
        'multiprocess': {
            'config': {
                'max_concurrent': 0,
                'retries': {
                    'disabled': {
                    },
                    'enabled': {
                    }
                }
            }
        }
    },
    'loggers': {
        'console': {
            'config': {
                'log_level': '',
                'name': ''
            }
        }
    },
    'resources': {
        'value': {
            'config': {
                'mode_two_field': 0
            }
        }
    },
    'solids': {
    },
    'storage': {
        'filesystem': {
            'config': {
                'base_dir': ''
            }
        },
        'in_memory': {
        }
    }
}



from dagster import RepositoryDefinition, pipeline, solid
from dagster.core.snap import PipelineSnapshot, active_repository_data_from_def
from dagster.core.snap.active_data import ActivePipelineData


def test_repository_snap_all_props():
    @solid
    def noop_solid(_):
        pass

    @pipeline
    def noop_pipeline():
        noop_solid()

    repo = RepositoryDefinition(name='noop_repo', pipeline_defs=[noop_pipeline])
    active_repo_data = active_repository_data_from_def(repo)

    assert active_repo_data.name == 'noop_repo'
    assert len(active_repo_data.active_pipeline_datas) == 1
    assert isinstance(active_repo_data.active_pipeline_datas[0], ActivePipelineData)

    pipeline_snapshot = active_repo_data.active_pipeline_datas[0].pipeline_snapshot
    assert isinstance(pipeline_snapshot, PipelineSnapshot)
    assert pipeline_snapshot.name == 'noop_pipeline'
    assert pipeline_snapshot.description is None
    assert pipeline_snapshot.tags == {}


def test_repository_snap_empty():
    repo = RepositoryDefinition(name='empty_repo', pipeline_defs=[])
    active_repo_data = active_repository_data_from_def(repo)
    assert active_repo_data.name == 'empty_repo'
    assert len(active_repo_data.active_pipeline_datas) == 0



from dagster import PipelineDefinition, execute_pipeline, pipeline, solid
from dagster.core.execution.api import create_execution_plan
from dagster.core.instance import DagsterInstance
from dagster.core.snap import (
    create_execution_plan_snapshot_id,
    create_pipeline_snapshot_id,
    snapshot_from_execution_plan,
)


def test_get_run_by_id():
    instance = DagsterInstance.ephemeral()

    assert instance.get_runs() == []
    pipeline_run = instance.create_run(
        pipeline_name='foo_pipeline', run_id='new_run', pipeline_snapshot=None
    )

    assert instance.get_runs() == [pipeline_run]

    assert instance.get_run_by_id(pipeline_run.run_id) == pipeline_run


def do_test_single_write_read(instance):
    run_id = 'some_run_id'
    pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])
    instance.create_run_for_pipeline(pipeline_def=pipeline_def, run_id=run_id)
    run = instance.get_run_by_id(run_id)
    assert run.run_id == run_id
    assert run.pipeline_name == 'some_pipeline'
    assert list(instance.get_runs()) == [run]
    instance.wipe()
    assert list(instance.get_runs()) == []


def test_filesystem_persist_one_run(tmpdir):
    do_test_single_write_read(DagsterInstance.local_temp(str(tmpdir)))


def test_in_memory_persist_one_run():
    do_test_single_write_read(DagsterInstance.ephemeral())


def test_create_pipeline_snapshot():
    @solid
    def noop_solid(_):
        pass

    @pipeline
    def noop_pipeline():
        noop_solid()

    instance = DagsterInstance.local_temp()

    result = execute_pipeline(noop_pipeline, instance=instance)
    assert result.success

    run = instance.get_run_by_id(result.run_id)

    assert run.pipeline_snapshot_id == create_pipeline_snapshot_id(
        noop_pipeline.get_pipeline_snapshot()
    )


def test_create_execution_plan_snapshot():
    @solid
    def noop_solid(_):
        pass

    @pipeline
    def noop_pipeline():
        noop_solid()

    instance = DagsterInstance.local_temp()

    execution_plan = create_execution_plan(noop_pipeline)

    ep_snapshot = snapshot_from_execution_plan(
        execution_plan, noop_pipeline.get_pipeline_snapshot_id()
    )
    ep_snapshot_id = create_execution_plan_snapshot_id(ep_snapshot)

    result = execute_pipeline(noop_pipeline, instance=instance)
    assert result.success

    run = instance.get_run_by_id(result.run_id)

    assert run.execution_plan_snapshot_id == ep_snapshot_id
    assert run.execution_plan_snapshot_id == create_execution_plan_snapshot_id(ep_snapshot)



import pytest
import six

from dagster.utils.backoff import backoff, backoff_delay_generator


class UnretryableException(Exception):
    pass


class RetryableException(Exception):
    pass


class RetryableExceptionB(Exception):
    pass


class Failer:
    def __init__(self, fails=0, exception=RetryableException):
        self.fails = fails
        self.exception = exception
        self.call_count = 0
        self.args = []
        self.kwargs = []

    def __call__(self, *args, **kwargs):
        self.call_count += 1
        self.args.append(args)
        self.kwargs.append(kwargs)

        if self.call_count <= self.fails:
            raise self.exception

        return True


def test_backoff_delay_generator():
    gen = backoff_delay_generator()
    vals = []
    for _ in range(10):
        vals.append(six.next(gen))

    assert vals == [0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, 25.6, 51.2]


def test_backoff():
    fn = Failer(fails=100)
    with pytest.raises(RetryableException):
        backoff(fn, retry_on=(RetryableException,), args=[3, 2, 1], kwargs={'foo': 'bar'})

    assert fn.call_count == 5
    assert all([args == (3, 2, 1) for args in fn.args])
    assert all([kwargs == {'foo': 'bar'} for kwargs in fn.kwargs])

    fn = Failer()
    assert backoff(fn, retry_on=(RetryableException,), args=[3, 2, 1], kwargs={'foo': 'bar'})
    assert fn.call_count == 1

    fn = Failer(fails=1)
    assert backoff(fn, retry_on=(RetryableException,), args=[3, 2, 1], kwargs={'foo': 'bar'})
    assert fn.call_count == 2

    fn = Failer(fails=1)
    with pytest.raises(RetryableException):
        backoff(
            fn, retry_on=(RetryableException,), args=[3, 2, 1], kwargs={'foo': 'bar'}, max_retries=0
        )
    assert fn.call_count == 1

    fn = Failer(fails=2)
    with pytest.raises(RetryableException):
        backoff(
            fn, retry_on=(RetryableException,), args=[3, 2, 1], kwargs={'foo': 'bar'}, max_retries=1
        )
    assert fn.call_count == 2



import re

import pytest

from dagster import (
    DagsterInvariantViolationError,
    DagsterTypeCheckDidNotPass,
    DependencyDefinition,
    EventMetadataEntry,
    InputDefinition,
    Output,
    OutputDefinition,
    PipelineDefinition,
    SolidDefinition,
    check,
    execute_pipeline,
    execute_solid,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.core.test_utils import single_output_solid


def create_root_success_solid(name):
    def root_fn(_context, _args):
        passed_rows = []
        passed_rows.append({name: 'compute_called'})
        return passed_rows

    return single_output_solid(
        name=name, input_defs=[], compute_fn=root_fn, output_def=OutputDefinition()
    )


def create_root_fn_failure_solid(name):
    def failed_fn(**_kwargs):
        raise Exception('Compute failed')

    return single_output_solid(
        name=name, input_defs=[], compute_fn=failed_fn, output_def=OutputDefinition()
    )


def test_compute_failure_pipeline():
    pipeline_def = PipelineDefinition(solid_defs=[create_root_fn_failure_solid('failing')])
    pipeline_result = execute_pipeline(pipeline_def, raise_on_error=False)

    assert not pipeline_result.success

    result_list = pipeline_result.solid_result_list

    assert len(result_list) == 1
    assert not result_list[0].success
    assert result_list[0].failure_data


def test_failure_midstream():
    '''
    A
     \\
       C (fails) = D (skipped)
     //
    B
    '''

    solid_a = create_root_success_solid('A')
    solid_b = create_root_success_solid('B')

    def fail_fn(_context, inputs):
        check.failed('user error')
        return [inputs['A'], inputs['B'], {'C': 'compute_called'}]

    def success_fn(_context, inputs):
        return [inputs['C'], {'D': 'compute_called'}]

    solid_c = single_output_solid(
        name='C',
        input_defs=[InputDefinition(name='A'), InputDefinition(name='B')],
        compute_fn=fail_fn,
        output_def=OutputDefinition(),
    )

    solid_d = single_output_solid(
        name='D',
        input_defs=[InputDefinition(name='C')],
        compute_fn=success_fn,
        output_def=OutputDefinition(),
    )

    pipeline_def = PipelineDefinition(
        solid_defs=[solid_a, solid_b, solid_c, solid_d],
        dependencies={
            'C': {'A': DependencyDefinition(solid_a.name), 'B': DependencyDefinition(solid_b.name)},
            'D': {'C': DependencyDefinition(solid_c.name)},
        },
    )
    pipeline_result = execute_pipeline(pipeline_def, raise_on_error=False)

    assert pipeline_result.result_for_solid('A').success
    assert pipeline_result.result_for_solid('B').success
    assert not pipeline_result.result_for_solid('C').success
    assert pipeline_result.result_for_solid('C').failure_data.error.cls_name == 'CheckError'
    assert not pipeline_result.result_for_solid('D').success
    assert pipeline_result.result_for_solid('D').skipped


def test_failure_propagation():
    '''
      B =========== C
     //             \\
    A                F (skipped)
     \\             //
      D (fails) == E (skipped)
    '''

    solid_a = create_root_success_solid('A')

    def fail_fn(_context, inputs):
        check.failed('user error')
        return inputs

    def success_fn(_context, inputs):
        return inputs

    solid_b = single_output_solid(
        name='B',
        input_defs=[InputDefinition(name='A')],
        compute_fn=success_fn,
        output_def=OutputDefinition(),
    )

    solid_c = single_output_solid(
        name='C',
        input_defs=[InputDefinition(name='B')],
        compute_fn=success_fn,
        output_def=OutputDefinition(),
    )

    solid_d = single_output_solid(
        name='D',
        input_defs=[InputDefinition(name='A')],
        compute_fn=fail_fn,
        output_def=OutputDefinition(),
    )

    solid_e = single_output_solid(
        name='E',
        input_defs=[InputDefinition(name='D')],
        compute_fn=success_fn,
        output_def=OutputDefinition(),
    )

    solid_f = single_output_solid(
        name='F',
        input_defs=[InputDefinition(name='C'), InputDefinition(name='E')],
        compute_fn=success_fn,
        output_def=OutputDefinition(),
    )

    pipeline_def = PipelineDefinition(
        solid_defs=[solid_a, solid_b, solid_c, solid_d, solid_e, solid_f],
        dependencies={
            'B': {'A': DependencyDefinition(solid_a.name)},
            'D': {'A': DependencyDefinition(solid_a.name)},
            'C': {'B': DependencyDefinition(solid_b.name)},
            'E': {'D': DependencyDefinition(solid_d.name)},
            'F': {'C': DependencyDefinition(solid_c.name), 'E': DependencyDefinition(solid_e.name)},
        },
    )

    pipeline_result = execute_pipeline(pipeline_def, raise_on_error=False)

    assert pipeline_result.result_for_solid('A').success
    assert pipeline_result.result_for_solid('B').success
    assert pipeline_result.result_for_solid('C').success
    assert not pipeline_result.result_for_solid('D').success
    assert pipeline_result.result_for_solid('D').failure_data.error.cls_name == 'CheckError'
    assert not pipeline_result.result_for_solid('E').success
    assert pipeline_result.result_for_solid('E').skipped
    assert not pipeline_result.result_for_solid('F').success
    assert pipeline_result.result_for_solid('F').skipped


def test_do_not_yield_result():
    solid_inst = SolidDefinition(
        name='do_not_yield_result',
        input_defs=[],
        output_defs=[OutputDefinition()],
        compute_fn=lambda *_args, **_kwargs: Output('foo'),
    )

    with pytest.raises(
        DagsterInvariantViolationError,
        match='Compute function for solid do_not_yield_result returned a Output',
    ):
        execute_solid(solid_inst)


def test_yield_non_result():
    @solid
    def yield_wrong_thing(_):
        yield 'foo'

    with pytest.raises(
        DagsterInvariantViolationError,
        match=re.escape('Compute function for solid yield_wrong_thing yielded a value of type <')
        + r'(class|type)'
        + re.escape(
            ' \'str\'> rather than an instance of Output, Materialization, or ExpectationResult.'
        ),
    ):
        execute_solid(yield_wrong_thing)


def test_single_compute_fn_returning_result():
    solid_inst = single_output_solid(
        'test_return_result',
        input_defs=[],
        compute_fn=lambda *_args, **_kwargs: Output(None),
        output_def=OutputDefinition(),
    )

    with pytest.raises(DagsterInvariantViolationError):
        execute_solid(solid_inst)


def test_user_error_propogation():
    err_msg = 'the user has errored'

    class UserError(Exception):
        pass

    @lambda_solid
    def throws_user_error():
        raise UserError(err_msg)

    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(input_defs=[InputDefinition('num')])
    def add_one(num):
        return num + 1

    pipeline_def = PipelineDefinition(
        name='test_user_error_propogation',
        solid_defs=[throws_user_error, return_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )

    with pytest.raises(UserError) as e_info:
        execute_pipeline(pipeline_def)

    assert isinstance(e_info.value, UserError)


def test_explicit_failure():
    @lambda_solid
    def throws_failure():
        raise DagsterTypeCheckDidNotPass(
            description='Always fails.',
            metadata_entries=[EventMetadataEntry.text('why', label='always_fails')],
        )

    @pipeline
    def pipe():
        throws_failure()

    with pytest.raises(DagsterTypeCheckDidNotPass) as exc_info:
        execute_pipeline(pipe)

    assert exc_info.value.description == 'Always fails.'
    assert exc_info.value.metadata_entries == [EventMetadataEntry.text('why', label='always_fails')]



import abc

import six


class ScheduleStorage(six.with_metaclass(abc.ABCMeta)):
    '''Abstract class for managing persistance of scheduler artifacts
    '''

    @abc.abstractmethod
    def all_schedules(self, repository=None):
        '''Return all schedules present in the storage

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
        '''

    @abc.abstractmethod
    def get_schedule_by_name(self, repository, schedule_name):
        '''Return the unique schedule with the given name

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule_name (str): The name of the schedule
        '''

    @abc.abstractmethod
    def add_schedule(self, repository, schedule):
        '''Add a schedule to storage.

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule (Schedule): The schedule to add
        '''

    @abc.abstractmethod
    def update_schedule(self, repository, schedule):
        '''Update a schedule already in storage, using schedule name to match schedules.

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule (Schedule): The schedule to update
        '''

    @abc.abstractmethod
    def delete_schedule(self, repository, schedule):
        '''Delete a schedule from storage.

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule (Schedule): The schedule to delete
        '''

    @abc.abstractmethod
    def get_schedule_ticks_by_schedule(self, repository, schedule_name):
        '''Get all schedule ticks for a given schedule

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule_name (str): The name of the schedule
        '''

    @abc.abstractmethod
    def get_schedule_tick_stats_by_schedule(self, repository, schedule_name):
        '''Get summary statistics for schedule ticks for a given schedule

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule_name (str): The name of the schedule
        '''

    @abc.abstractmethod
    def create_schedule_tick(self, repository, schedule_tick_data):
        '''Add a schedule tick to storage.

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            schedule_tick_data (ScheduleTickData): The schedule tick to add
        '''

    @abc.abstractmethod
    def update_schedule_tick(self, repository, tick):
        '''Update a schedule tick already in storage.

        Args:
            repository (RepositoryDefinition): The repository the schedule belongs to
            tick (ScheduleTick): The schedule tick to update
        '''

    @abc.abstractmethod
    def wipe(self):
        '''Delete all schedules from storage
        '''



import os

import six

from dagster import check

from .config_type import ConfigIntInstance, ConfigStringInstance, ScalarUnion
from .errors import PostProcessingError
from .field_utils import Selector

VALID_STRING_SOURCE_TYPES = six.string_types + (dict,)


def _ensure_env_variable(var):
    check.str_param(var, 'var')
    value = os.getenv(var)
    if value is None:
        raise PostProcessingError(
            (
                'You have attempted to fetch the environment variable "{var}" '
                'which is not set. In order for this execution to succeed it '
                'must be set in this environment.'
            ).format(var=var)
        )
    return value


class StringSourceType(ScalarUnion):
    def __init__(self):
        super(StringSourceType, self).__init__(
            scalar_type=ConfigStringInstance,
            non_scalar_type=Selector({'env': str}),
            _key='StringSourceType',
        )

    def post_process(self, value):
        check.param_invariant(isinstance(value, VALID_STRING_SOURCE_TYPES), 'value')

        if not isinstance(value, dict):
            return value

        key, cfg = list(value.items())[0]
        check.invariant(key == 'env', 'Only valid key is env')
        return str(_ensure_env_variable(cfg))


class IntSourceType(ScalarUnion):
    def __init__(self):
        super(IntSourceType, self).__init__(
            scalar_type=ConfigIntInstance,
            non_scalar_type=Selector({'env': str}),
            _key='IntSourceType',
        )

    def post_process(self, value):
        check.param_invariant(isinstance(value, (dict, int)), 'value', 'Should be pre-validated')

        if not isinstance(value, dict):
            return value

        check.invariant(len(value) == 1, 'Selector should have one entry')

        key, cfg = list(value.items())[0]
        check.invariant(key == 'env', 'Only valid key is env')
        value = _ensure_env_variable(cfg)
        try:
            return int(value)
        except ValueError:
            raise PostProcessingError(
                (
                    'Value "{value}" stored in env variable "{var}" cannot be '
                    'coerced into an int.'
                ).format(value=value, var=cfg)
            )


StringSource = StringSourceType()
IntSource = IntSourceType()



import abc

import six


class Engine(six.with_metaclass(abc.ABCMeta)):  # pylint: disable=no-init
    '''
    The Engine abstraction exists to allow DAG execution to be broken up and distrubted in different ways.
    Each unique engine should map to some system for managing work, such as using multiple processes on a single machine or
    using a task queue service to distrbute work across remote machines.
    '''

    @staticmethod
    @abc.abstractmethod
    def execute(pipeline_context, execution_plan):
        '''
        For the given context and execution plan, orchestrate a series of sub plan executions in a way that satisfies the whole plan being executed.

        Args:
            pipeline_context (SystemPipelineExecutionContext): The pipeline execution context.
            execution_plan (ExecutionPlan): The plan to execute.
        '''



from dagster import ModeDefinition, PresetDefinition, RepositoryDefinition, pipeline, solid
from dagster.core.snap import active_repository_data_from_def
from dagster.core.snap.active_data import active_pipeline_data_from_def
from dagster.serdes import serialize_pp


@solid
def a_solid(_):
    pass


@pipeline(
    mode_defs=[ModeDefinition('default'), ModeDefinition('mode_one')],
    preset_defs=[
        PresetDefinition(name='plain_preset'),
        PresetDefinition(
            name='kitchen_sink_preset',
            environment_dict={'foo': 'bar'},
            solid_subset=['a_solid'],
            mode='mode_one',
        ),
    ],
)
def a_pipeline():
    a_solid()


def test_active_repository_data(snapshot):
    rep_def = RepositoryDefinition(name='repo', pipeline_defs=[a_pipeline])
    snapshot.assert_match(serialize_pp(active_repository_data_from_def(rep_def)))


def test_active_pipeline_data(snapshot):
    snapshot.assert_match(serialize_pp(active_pipeline_data_from_def(a_pipeline)))



from dagster import (
    ExecutionTargetHandle,
    PartitionSetDefinition,
    RepositoryDefinition,
    ScheduleDefinition,
    lambda_solid,
    pipeline,
    repository_partitions,
    schedules,
)
from dagster.utils import file_relative_path


def test_deprecated_load():
    repo = ExecutionTargetHandle.for_repo_yaml(
        file_relative_path(__file__, 'repository.yaml')
    ).build_repository_definition()
    assert len(repo.get_all_pipelines()) == 1
    assert len(repo.schedule_defs) == 1
    assert len(repo.partition_set_defs) == 1


def define_repository():
    @lambda_solid
    def hello():
        return 'hello'

    @pipeline
    def test():
        hello()

    return RepositoryDefinition('test', pipeline_defs=[test])


@schedules
def define_schedules():
    return [ScheduleDefinition('test', '0 0 * * *', 'test')]


@repository_partitions
def define_partitions():
    return [
        PartitionSetDefinition(
            name="test",
            pipeline_name="test",
            partition_fn=lambda: ["test"],
            environment_dict_fn_for_partition=lambda _: {},
        )
    ]



from collections import OrderedDict, namedtuple

from dagster import check
from dagster.core.definitions import PipelineDefinition, PresetDefinition, RepositoryDefinition
from dagster.serdes import whitelist_for_serdes

from .pipeline_snapshot import PipelineIndex, PipelineSnapshot


class RepositoryIndex:
    def __init__(self, active_repository_data):
        self.active_repository_data = check.inst_param(
            active_repository_data, 'active_repository_data', ActiveRepositoryData
        )
        self._pipeline_index_map = OrderedDict(
            (
                active_pipeline_data.pipeline_snapshot.name,
                PipelineIndex(active_pipeline_data.pipeline_snapshot),
            )
            for active_pipeline_data in active_repository_data.active_pipeline_datas
        )

    def get_pipeline_index(self, pipeline_name):
        return self._pipeline_index_map[pipeline_name]

    def has_pipeline(self, pipeline_name):
        return pipeline_name in self._pipeline_index_map

    def get_pipeline_indices(self):
        return self._pipeline_index_map.values()

    @staticmethod
    def from_repository_def(repository_definition):
        return RepositoryIndex(active_repository_data_from_def(repository_definition))


@whitelist_for_serdes
class ActiveRepositoryData(namedtuple('_ActiveRepositoryData', 'name active_pipeline_datas')):
    def __new__(cls, name, active_pipeline_datas):
        return super(ActiveRepositoryData, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            active_pipeline_datas=check.list_param(
                active_pipeline_datas, 'active_pipeline_datas', of_type=ActivePipelineData
            ),
        )

    def get_pipeline_snapshot(self, name):
        check.str_param(name, 'name')

        for active_pipeline_data in self.active_pipeline_datas:
            if active_pipeline_data.name == name:
                return active_pipeline_data.pipeline_snapshot

        check.failed('Could not find pipeline snapshot named ' + name)

    def get_active_pipeline_data(self, name):
        check.str_param(name, 'name')

        for active_pipeline in self.active_pipeline_datas:
            if active_pipeline.name == name:
                return active_pipeline

        check.failed('Could not find active pipeline data named ' + name)


@whitelist_for_serdes
class ActivePipelineData(
    namedtuple('_ActivePipelineData', 'name pipeline_snapshot active_presets')
):
    def __new__(cls, name, pipeline_snapshot, active_presets):
        return super(ActivePipelineData, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            pipeline_snapshot=check.inst_param(
                pipeline_snapshot, 'pipeline_snapshot', PipelineSnapshot
            ),
            active_presets=check.list_param(
                active_presets, 'active_presets', of_type=ActivePresetData
            ),
        )


@whitelist_for_serdes
class ActivePresetData(namedtuple('_ActivePresetData', 'name environment_dict solid_subset mode')):
    def __new__(cls, name, environment_dict, solid_subset, mode):
        return super(ActivePresetData, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            environment_dict=check.opt_dict_param(environment_dict, 'environment_dict'),
            solid_subset=check.list_param(solid_subset, 'solid_subset', of_type=str)
            if solid_subset is not None
            else None,
            mode=check.str_param(mode, 'mode'),
        )


def active_repository_data_from_def(repository_def):
    check.inst_param(repository_def, 'repository_def', RepositoryDefinition)

    return ActiveRepositoryData(
        name=repository_def.name,
        active_pipeline_datas=sorted(
            list(map(active_pipeline_data_from_def, repository_def.get_all_pipelines())),
            key=lambda pd: pd.name,
        ),
    )


def active_pipeline_data_from_def(pipeline_def):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    return ActivePipelineData(
        name=pipeline_def.name,
        pipeline_snapshot=PipelineSnapshot.from_pipeline_def(pipeline_def),
        active_presets=sorted(
            list(map(active_preset_data_from_def, pipeline_def.preset_defs)), key=lambda pd: pd.name
        ),
    )


def active_preset_data_from_def(preset_def):
    check.inst_param(preset_def, 'preset_def', PresetDefinition)
    return ActivePresetData(
        name=preset_def.name,
        environment_dict=preset_def.environment_dict,
        solid_subset=preset_def.solid_subset,
        mode=preset_def.mode,
    )



import os

from dagster import check
from dagster.core.events import DagsterEvent, EngineEventData
from dagster.core.execution.config import ExecutorConfig
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.execute_plan import inner_plan_execution_iterator
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.utils.timing import format_duration, time_execution_scope

from .engine_base import Engine


class InProcessEngine(Engine):  # pylint: disable=no-init
    @staticmethod
    def execute(pipeline_context, execution_plan):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)

        step_keys_to_execute = execution_plan.step_keys_to_execute

        yield DagsterEvent.engine_event(
            pipeline_context,
            'Executing steps in process (pid: {pid})'.format(pid=os.getpid()),
            event_specific_data=EngineEventData.in_process(os.getpid(), step_keys_to_execute,),
        )

        with time_execution_scope() as timer_result:
            check.param_invariant(
                isinstance(pipeline_context.executor_config, ExecutorConfig),
                'pipeline_context',
                'Expected executor_config to be ExecutorConfig got {}'.format(
                    pipeline_context.executor_config
                ),
            )

            for event in inner_plan_execution_iterator(
                pipeline_context, execution_plan, pipeline_context.executor_config.retries
            ):
                yield event

        yield DagsterEvent.engine_event(
            pipeline_context,
            'Finished steps in process (pid: {pid}) in {duration_ms}'.format(
                pid=os.getpid(), duration_ms=format_duration(timer_result.millis)
            ),
            event_specific_data=EngineEventData.in_process(os.getpid(), step_keys_to_execute),
        )



from dagster import DagsterInstance, ExecutionTargetHandle, execute_pipeline, lambda_solid, pipeline


def define_pipeline():
    @lambda_solid
    def ping():
        return 'ping'

    @pipeline
    def simple():
        ping()

    return simple


def test_multiproc_markers():
    pipe = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_pipeline'
    ).build_pipeline_definition()
    instance = DagsterInstance.local_temp()
    result = execute_pipeline(
        pipe,
        instance=instance,
        environment_dict={'execution': {'multiprocess': {}}, 'storage': {'filesystem': {}}},
    )
    assert result.success
    events = instance.all_logs(result.run_id)
    start_markers = {}
    end_markers = {}
    for event in events:
        dagster_event = event.dagster_event
        if dagster_event.is_engine_event:
            if dagster_event.engine_event_data.marker_start:
                key = '{step}.{marker}'.format(
                    step=event.step_key, marker=dagster_event.engine_event_data.marker_start
                )
                start_markers[key] = event.timestamp
            if dagster_event.engine_event_data.marker_end:
                key = '{step}.{marker}'.format(
                    step=event.step_key, marker=dagster_event.engine_event_data.marker_end
                )
                end_markers[key] = event.timestamp

    seen = set()
    assert set(start_markers.keys()) == set(end_markers.keys())
    for key in end_markers:
        assert end_markers[key] - start_markers[key] > 0
        seen.add(key)

    assert 'ping.compute.multiprocess_subprocess_init' in end_markers



import pytest

from dagster import (
    Any,
    Bool,
    DagsterInvalidDefinitionError,
    Dict,
    Float,
    InputDefinition,
    Int,
    List,
    Nothing,
    Optional,
    OutputDefinition,
    Path,
    PipelineDefinition,
    Set,
    String,
    Tuple,
    lambda_solid,
    pipeline,
)
from dagster.core.types.dagster_type import (
    ALL_RUNTIME_BUILTINS,
    DagsterType,
    DagsterTypeKind,
    resolve_dagster_type,
)


def inner_type_key_set(dagster_type):
    return {t.key for t in dagster_type.inner_types}


def test_inner_types():
    assert resolve_dagster_type(Int).inner_types == []

    list_int_runtime = resolve_dagster_type(List[Int])
    assert inner_type_key_set(list_int_runtime) == set(['Int'])

    list_list_int_runtime = resolve_dagster_type(List[List[Int]])
    assert inner_type_key_set(list_list_int_runtime) == set(['Int', 'List.Int'])

    list_nullable_int_runtime = resolve_dagster_type(List[Optional[Int]])
    assert inner_type_key_set(list_nullable_int_runtime) == set(['Int', 'Optional.Int'])
    assert not list_nullable_int_runtime.kind == DagsterTypeKind.SCALAR


def test_is_any():
    assert not resolve_dagster_type(Int).kind == DagsterTypeKind.ANY
    assert resolve_dagster_type(Int).kind == DagsterTypeKind.SCALAR


def test_display_name():

    int_runtime = resolve_dagster_type(Int)
    assert int_runtime.display_name == 'Int'
    list_int_runtime = resolve_dagster_type(List[Int])
    assert list_int_runtime.display_name == '[Int]'
    list_list_int_runtime = resolve_dagster_type(List[List[Int]])
    assert list_list_int_runtime.display_name == '[[Int]]'
    list_nullable_int_runtime = resolve_dagster_type(List[Optional[Int]])
    assert list_nullable_int_runtime.display_name == '[Int?]'


def test_builtins_available():
    pipeline_def = PipelineDefinition(name='test_builting_available', solid_defs=[])
    for builtin_type in ALL_RUNTIME_BUILTINS:
        assert pipeline_def.has_dagster_type(builtin_type.name)
        assert pipeline_def.dagster_type_named(builtin_type.name).is_builtin


def test_python_mapping():
    runtime = resolve_dagster_type(int)
    assert runtime.name == 'Int'
    runtime = resolve_dagster_type(str)
    assert runtime.name == 'String'
    runtime = resolve_dagster_type(bool)
    assert runtime.name == 'Bool'
    runtime = resolve_dagster_type(float)
    assert runtime.name == 'Float'

    @lambda_solid(input_defs=[InputDefinition('num', int)])
    def add_one(num):
        return num + 1

    assert add_one.input_defs[0].dagster_type.name == 'Int'

    runtime = resolve_dagster_type(float)
    runtime.type_check(None, 1.0)
    res = runtime.type_check(None, 1)
    assert not res.success

    runtime = resolve_dagster_type(bool)
    runtime.type_check(None, True)
    res = runtime.type_check(None, 1)
    assert not res.success


def test_double_dagster_type():

    AlwaysSucceedsFoo = DagsterType(name='Foo', type_check_fn=lambda _, _val: True)
    AlwaysFailsFoo = DagsterType(name='Foo', type_check_fn=lambda _, _val: False)

    @lambda_solid
    def return_a_thing():
        return 1

    @lambda_solid(
        input_defs=[InputDefinition('succeeds', AlwaysSucceedsFoo)],
        output_def=OutputDefinition(AlwaysFailsFoo),
    )
    def yup(succeeds):
        return succeeds

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:

        @pipeline
        def _should_fail():
            yup(return_a_thing())

    assert str(exc_info.value) == (
        'You have created two dagster types with the same name "Foo". '
        'Dagster types have must have unique names.'
    )


def test_comparison():
    # Base types
    assert resolve_dagster_type(Any) == resolve_dagster_type(Any)
    assert resolve_dagster_type(String) == resolve_dagster_type(String)
    assert resolve_dagster_type(Bool) == resolve_dagster_type(Bool)
    assert resolve_dagster_type(Float) == resolve_dagster_type(Float)
    assert resolve_dagster_type(Int) == resolve_dagster_type(Int)
    assert resolve_dagster_type(Path) == resolve_dagster_type(Path)
    assert resolve_dagster_type(String) == resolve_dagster_type(String)
    assert resolve_dagster_type(Nothing) == resolve_dagster_type(Nothing)
    assert resolve_dagster_type(Optional[String]) == resolve_dagster_type(Optional[String])

    types = [Any, Bool, Float, Int, String, Nothing]
    non_equal_pairs = [(t1, t2) for t1 in types for t2 in types if t1 != t2]
    for t1, t2 in non_equal_pairs:
        assert resolve_dagster_type(t1) != resolve_dagster_type(t2)
    assert resolve_dagster_type(Optional[String]) != resolve_dagster_type(Optional[Int])

    # List type
    assert resolve_dagster_type(List) == resolve_dagster_type(List)
    assert resolve_dagster_type(List[String]) == resolve_dagster_type(List[String])
    assert resolve_dagster_type(List[List[Int]]) == resolve_dagster_type(List[List[Int]])
    assert resolve_dagster_type(List[Optional[String]]) == resolve_dagster_type(
        List[Optional[String]]
    )

    assert resolve_dagster_type(List[String]) != resolve_dagster_type(List[Int])
    assert resolve_dagster_type(List[List[String]]) != resolve_dagster_type(List[List[Int]])
    assert resolve_dagster_type(List[String]) != resolve_dagster_type(List[Optional[String]])

    # Tuple type
    assert resolve_dagster_type(Tuple) == resolve_dagster_type(Tuple)
    assert resolve_dagster_type(Tuple[String, Int]) == resolve_dagster_type(Tuple[String, Int])
    assert resolve_dagster_type(Tuple[Tuple[String, Int]]) == resolve_dagster_type(
        Tuple[Tuple[String, Int]]
    )
    assert resolve_dagster_type(Tuple[Optional[String], Int]) == resolve_dagster_type(
        Tuple[Optional[String], Int]
    )

    assert resolve_dagster_type(Tuple[String, Int]) != resolve_dagster_type(Tuple[Int, String])
    assert resolve_dagster_type(Tuple[Tuple[String, Int]]) != resolve_dagster_type(
        Tuple[Tuple[Int, String]]
    )
    assert resolve_dagster_type(Tuple[String]) != resolve_dagster_type(Tuple[Optional[String]])

    # Set type
    assert resolve_dagster_type(Set) == resolve_dagster_type(Set)
    assert resolve_dagster_type(Set[String]) == resolve_dagster_type(Set[String])
    assert resolve_dagster_type(Set[Set[Int]]) == resolve_dagster_type(Set[Set[Int]])
    assert resolve_dagster_type(Set[Optional[String]]) == resolve_dagster_type(
        Set[Optional[String]]
    )

    assert resolve_dagster_type(Set[String]) != resolve_dagster_type(Set[Int])
    assert resolve_dagster_type(Set[Set[String]]) != resolve_dagster_type(Set[Set[Int]])
    assert resolve_dagster_type(Set[String]) != resolve_dagster_type(Set[Optional[String]])

    # Dict type
    assert resolve_dagster_type(Dict) == resolve_dagster_type(Dict)
    assert resolve_dagster_type(Dict[String, Int]) == resolve_dagster_type(Dict[String, Int])
    assert resolve_dagster_type(Dict[String, Dict[String, Int]]) == resolve_dagster_type(
        Dict[String, Dict[String, Int]]
    )

    assert resolve_dagster_type(Dict[String, Int]) != resolve_dagster_type(Dict[Int, String])
    assert resolve_dagster_type(Dict[Int, Dict[String, Int]]) != resolve_dagster_type(
        Dict[String, Dict[String, Int]]
    )



import pytest
from click.testing import CliRunner

from dagster import file_relative_path
from dagster.check import ParameterCheckError
from dagster.cli.repository import snapshot_command
from dagster.core.snap import ActiveRepositoryData
from dagster.serdes import deserialize_json_to_dagster_namedtuple
from dagster.utils import safe_tempfile_path


def test_snapshot_command_handle_repository():
    runner = CliRunner()
    with safe_tempfile_path() as fp:
        result = runner.invoke(
            snapshot_command, [fp, '-y', file_relative_path(__file__, 'repository_file.yaml')],
        )
        assert result.exit_code == 0
        # Now that we have the snapshot make sure that it can be properly deserialized
        with open(fp) as buffer:
            active_repository_data = deserialize_json_to_dagster_namedtuple(buffer.read())
        assert isinstance(active_repository_data, ActiveRepositoryData)
        assert active_repository_data.name == 'bar'
        assert len(active_repository_data.active_pipeline_datas) == 2


def test_snapshot_command_error_on_pipeline_definition():
    runner = CliRunner()
    with pytest.raises(ParameterCheckError):
        with safe_tempfile_path() as fp:
            result = runner.invoke(
                snapshot_command,
                [
                    fp,
                    '-f',
                    file_relative_path(__file__, 'test_cli_commands.py'),
                    '-n',
                    'baz_pipeline',
                ],
            )
            assert result.exit_code == 1
            raise result.exception



import os

import pytest
import yaml

from dagster import check
from dagster.utils import file_relative_path
from dagster.utils.yaml_utils import (
    load_yaml_from_glob_list,
    load_yaml_from_globs,
    load_yaml_from_path,
    merge_yaml_strings,
    merge_yamls,
)


def test_load_yaml():
    assert load_yaml_from_path(file_relative_path(__file__, 'yamls/yaml_one.yaml')) == {
        'key_one': {'key_one_one': 'value_one'}
    }


def test_from_glob_list():
    assert load_yaml_from_glob_list([file_relative_path(__file__, 'yamls/yaml_one.yaml')]) == {
        'key_one': {'key_one_one': 'value_one'}
    }

    assert load_yaml_from_glob_list(
        [
            file_relative_path(__file__, 'yamls/yaml_one.yaml'),
            file_relative_path(__file__, 'yamls/yaml_two.yaml'),
        ]
    ) == {'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}}

    assert load_yaml_from_glob_list([file_relative_path(__file__, 'yamls/*.yaml')]) == {
        'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}
    }

    assert load_yaml_from_globs(
        file_relative_path(__file__, 'yamls/yaml_one.yaml'),
        file_relative_path(__file__, 'yamls/yaml_two.yaml'),
    ) == {'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}}

    assert load_yaml_from_glob_list(['flskhfhjsdf']) == {}


def test_merge_yamls():
    assert merge_yamls(
        [
            file_relative_path(__file__, os.path.join('yamls', 'yaml_one.yaml')),
            file_relative_path(__file__, os.path.join('yamls', 'yaml_two.yaml')),
        ]
    ) == {'key_one': {'key_one_one': 'value_one', 'key_one_two': 'value_two'}}

    with pytest.raises(
        check.CheckError,
        match=(
            'Expected YAML from file .* to parse to dictionary, '
            'instead got: "this is a valid YAML string but not a dictionary"'
        ),
    ):
        merge_yamls(
            [
                file_relative_path(__file__, os.path.join('yamls', 'yaml_one.yaml')),
                file_relative_path(__file__, os.path.join('yamls', 'bad', 'a_string.yaml')),
            ]
        )


def test_merge_yaml_strings():
    a = '''
foo:
  bar: 1
baz: 3
'''
    b = '''
foo:
  one: "one"
other: 4
'''
    c = '''
final: "result"
'''

    assert merge_yaml_strings([a, b, c]) == {
        'final': 'result',
        'foo': {'one': 'one', 'bar': 1},
        'other': 4,
        'baz': 3,
    }

    override = '''final: "some other result"'''
    assert merge_yaml_strings([a, b, c, override]) == {
        'foo': {'bar': 1, 'one': 'one'},
        'baz': 3,
        'other': 4,
        'final': 'some other result',
    }

    string_yaml = 'this is a valid YAML string but not a dictionary'
    expected = 'Expected YAML dictionary, instead got: "{string_yaml}"'.format(
        string_yaml=string_yaml
    )

    with pytest.raises(check.CheckError, match=expected):
        merge_yaml_strings([a, string_yaml])

    with pytest.raises(
        yaml.YAMLError,
        match="while scanning for the next token\nfound character '`' that cannot start any token",
    ):
        bad_yaml = '--- `'
        merge_yaml_strings([a, bad_yaml])



from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.types.dagster_type import (
    BuiltinScalarDagsterType,
    DagsterType,
    resolve_dagster_type,
)
from dagster.utils.backcompat import rename_warning

from .utils import check_valid_name


class _NoValueSentinel(object):
    pass


# unfortunately since type_check functions need TypeCheckContext which is only available
# at runtime, we can only check basic types before runtime
def _check_default_value(input_name, dagster_type, default_value):
    if default_value is not _NoValueSentinel:
        if dagster_type.is_nothing:
            raise DagsterInvalidDefinitionError(
                'Setting a default_value is invalid on InputDefinitions of type Nothing'
            )

        if isinstance(dagster_type, BuiltinScalarDagsterType):
            type_check = dagster_type.type_check_scalar_value(default_value)
            if not type_check.success:
                raise DagsterInvalidDefinitionError(
                    (
                        'Type check failed for the default_value of InputDefinition '
                        '{input_name} of type {dagster_type}. '
                        'Received value {value} of type {type}'
                    ).format(
                        input_name=input_name,
                        dagster_type=dagster_type.name,
                        value=default_value,
                        type=type(default_value),
                    ),
                )

    return default_value


class InputDefinition(object):
    '''Defines an argument to a solid's compute function.

    Inputs may flow from previous solids' outputs, or be stubbed using config. They may optionally
    be typed using the Dagster type system.

    Args:
        name (str): Name of the input.
        dagster_type (Optional[Any]):  The type of this input. Users should provide one of the
            :ref:`built-in types <builtin>`, a dagster type explicitly constructed with
            :py:func:`as_dagster_type`, :py:func:`@usable_as_dagster_type <dagster_type`, or
            :py:func:`PythonObjectDagsterType`, or a Python type. Defaults to :py:class:`Any`.
        description (Optional[str]): Human-readable description of the input.
        default_value (Optional[Any]): The default value to use if no input is provided.
    '''

    def __init__(self, name, dagster_type=None, description=None, default_value=_NoValueSentinel):
        ''
        self._name = check_valid_name(name)

        self._dagster_type = check.inst(resolve_dagster_type(dagster_type), DagsterType)

        self._description = check.opt_str_param(description, 'description')

        self._default_value = _check_default_value(self._name, self._dagster_type, default_value)

    @property
    def name(self):
        return self._name

    @property
    def runtime_type(self):
        rename_warning(new_name='dagster_type', old_name='runtime_type', breaking_version='0.8.0')
        return self._dagster_type

    @property
    def dagster_type(self):
        return self._dagster_type

    @property
    def description(self):
        return self._description

    @property
    def has_default_value(self):
        return self._default_value is not _NoValueSentinel

    @property
    def default_value(self):
        check.invariant(self.has_default_value, 'Can only fetch default_value if has_default_value')
        return self._default_value

    def mapping_to(self, solid_name, input_name):
        '''Create an input mapping to an input of a child solid.

        In a CompositeSolidDefinition, you can use this helper function to construct
        an :py:class:`InputMapping` to the input of a child solid.

        Args:
            solid_name (str): The name of the child solid to which to map this input.
            input_name (str): The name of the child solid' input to which to map this input.

        Examples:

            .. code-block:: python

                input_mapping = InputDefinition('composite_input', Int).mapping_to(
                    'child_solid', 'int_input'
                )
        '''
        return InputMapping(self, solid_name, input_name)


class InputMapping(namedtuple('_InputMapping', 'definition solid_name input_name')):
    '''Defines an input mapping for a composite solid.

    Args:
        definition (InputDefinition): Defines the input to the composite solid.
        solid_name (str): The name of the child solid onto which to map the input.
        input_name (str): The name of the input to the child solid onto which to map the input.
    '''

    def __new__(cls, definition, solid_name, input_name):
        return super(InputMapping, cls).__new__(
            cls,
            check.inst_param(definition, 'definition', InputDefinition),
            check.str_param(solid_name, 'solid_name'),
            check.str_param(input_name, 'input_name'),
        )



from .config_type import ConfigScalar, ConfigScalarKind, ConfigType, Enum, EnumValue
from .errors import PostProcessingError
from .field import Field
from .field_utils import Permissive, Selector, Shape



import datetime
import logging
import os
import time
from abc import ABCMeta
from collections import defaultdict, namedtuple
from enum import Enum

import six
import yaml
from rx import Observable

from dagster import check, seven
from dagster.config import Field, Permissive
from dagster.core.definitions.pipeline import ExecutionSelector, PipelineDefinition
from dagster.core.errors import (
    DagsterInvalidConfigError,
    DagsterInvariantViolationError,
    DagsterRunAlreadyExists,
    DagsterRunConflict,
)
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus, PipelineRunsFilter
from dagster.serdes import ConfigurableClass, whitelist_for_serdes
from dagster.seven import get_current_datetime_in_utc
from dagster.utils.merger import merge_dicts
from dagster.utils.yaml_utils import load_yaml_from_globs

from .config import DAGSTER_CONFIG_YAML_FILENAME
from .ref import InstanceRef, compute_logs_directory

# 'airflow_execution_date' and 'is_airflow_ingest_pipeline' are hardcoded tags used in the
# airflow ingestion logic (see: dagster_pipeline_factory.py). 'airflow_execution_date' stores the
# 'execution_date' used in Airflow operator execution and 'is_airflow_ingest_pipeline' determines
# whether 'airflow_execution_date' is needed.
AIRFLOW_EXECUTION_DATE_STR = 'airflow_execution_date'
IS_AIRFLOW_INGEST_PIPELINE_STR = 'is_airflow_ingest_pipeline'


def _is_dagster_home_set():
    return bool(os.getenv('DAGSTER_HOME'))


def _dagster_home():
    dagster_home_path = os.getenv('DAGSTER_HOME')

    if not dagster_home_path:
        raise DagsterInvariantViolationError(
            'DAGSTER_HOME is not set, check is_dagster_home_set before invoking.'
        )

    return os.path.expanduser(dagster_home_path)


def _check_run_equality(pipeline_run, candidate_run):
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(candidate_run, 'candidate_run', PipelineRun)

    field_diff = {}
    for field in pipeline_run._fields:
        expected_value = getattr(pipeline_run, field)
        candidate_value = getattr(candidate_run, field)
        if expected_value != candidate_value:
            field_diff[field] = (expected_value, candidate_value)

    return field_diff


def _format_field_diff(field_diff):
    return '\n'.join(
        [
            (
                '    {field_name}:\n'
                + '        Expected: {expected_value}\n'
                + '        Received: {candidate_value}'
            ).format(
                field_name=field_name,
                expected_value=expected_value,
                candidate_value=candidate_value,
            )
            for field_name, (expected_value, candidate_value,) in field_diff.items()
        ]
    )


class _EventListenerLogHandler(logging.Handler):
    def __init__(self, instance):
        self._instance = instance
        super(_EventListenerLogHandler, self).__init__()

    def emit(self, record):
        from dagster.core.events.log import construct_event_record, StructuredLoggerMessage

        try:
            event = construct_event_record(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            )

            self._instance.handle_new_event(event)

        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during instance event listen')
            logging.exception(str(e))
            raise


class InstanceType(Enum):
    PERSISTENT = 'PERSISTENT'
    EPHEMERAL = 'EPHEMERAL'


class DagsterInstance:
    '''Core abstraction for managing Dagster's access to storage and other resources.

    Users should not directly instantiate this class; it is instantiated by internal machinery when
    ``dagit`` and ``dagster-graphql`` load, based on the values in the ``dagster.yaml`` file in
    ``$DAGSTER_HOME``. Configuration of this class should be done by setting values in that file.
    For example, to use Postgres for run and event log storage, you can write a ``dagster.yaml``
    such as the following:

    .. literalinclude:: ../../../../docs/sections/deploying/postgres_dagster.yaml
       :caption: dagster.yaml
       :language: YAML

    Args:
        instance_type (InstanceType): Indicates whether the instance is ephemeral or persistent.
            Users should not attempt to set this value directly or in their ``dagster.yaml`` files.
        local_artifact_storage (LocalArtifactStorage): The local artifact storage is used to
            configure storage for any artifacts that require a local disk, such as schedules, or
            when using the filesystem system storage to manage files and intermediates. By default,
            this will be a :py:class:`dagster.core.storage.root.LocalArtifactStorage`. Configurable
            in ``dagster.yaml`` using the :py:class:`~dagster.serdes.ConfigurableClass`
            machinery.
        run_storage (RunStorage): The run storage is used to store metadata about ongoing and past
            pipeline runs. By default, this will be a
            :py:class:`dagster.core.storage.runs.SqliteRunStorage`. Configurable in ``dagster.yaml``
            using the :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        event_storage (EventLogStorage): Used to store the structured event logs generated by
            pipeline runs. By default, this will be a
            :py:class:`dagster.core.storage.event_log.SqliteEventLogStorage`. Configurable in
            ``dagster.yaml`` using the :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        compute_log_manager (ComputeLogManager): The compute log manager handles stdout and stderr
            logging for solid compute functions. By default, this will be a
            :py:class:`dagster.core.storage.local_compute_log_manager.LocalComputeLogManager`.
            Configurable in ``dagster.yaml`` using the
            :py:class:`~dagster.serdes.ConfigurableClass` machinery.
        run_launcher (Optional[RunLauncher]): Optionally, a run launcher may be used to enable
            a Dagster instance to launch pipeline runs, e.g. on a remote Kubernetes cluster, in
            addition to running them locally.
        dagit_settings (Optional[Dict]): Specifies certain Dagit-specific, per-instance settings,
            such as feature flags. These are set in the ``dagster.yaml`` under the key ``dagit``.
        telemetry_settings (Optional[Dict]): Specifies certain telemetry-specific, per-instance
            settings, such as whether it is enabled. These are set in the ``dagster.yaml`` under
            the key ``telemetry``
        ref (Optional[InstanceRef]): Used by internal machinery to pass instances across process
            boundaries.
    '''

    _PROCESS_TEMPDIR = None

    def __init__(
        self,
        instance_type,
        local_artifact_storage,
        run_storage,
        event_storage,
        compute_log_manager,
        schedule_storage=None,
        scheduler=None,
        run_launcher=None,
        dagit_settings=None,
        telemetry_settings=None,
        ref=None,
    ):
        from dagster.core.storage.compute_log_manager import ComputeLogManager
        from dagster.core.storage.event_log import EventLogStorage
        from dagster.core.storage.root import LocalArtifactStorage
        from dagster.core.storage.runs import RunStorage
        from dagster.core.storage.schedules import ScheduleStorage
        from dagster.core.scheduler import Scheduler
        from dagster.core.launcher import RunLauncher

        self._instance_type = check.inst_param(instance_type, 'instance_type', InstanceType)
        self._local_artifact_storage = check.inst_param(
            local_artifact_storage, 'local_artifact_storage', LocalArtifactStorage
        )
        self._event_storage = check.inst_param(event_storage, 'event_storage', EventLogStorage)
        self._run_storage = check.inst_param(run_storage, 'run_storage', RunStorage)
        self._compute_log_manager = check.inst_param(
            compute_log_manager, 'compute_log_manager', ComputeLogManager
        )
        self._schedule_storage = check.opt_inst_param(
            schedule_storage, 'schedule_storage', ScheduleStorage
        )
        self._scheduler = check.opt_inst_param(scheduler, 'scheduler', Scheduler)
        self._run_launcher = check.opt_inst_param(run_launcher, 'run_launcher', RunLauncher)
        self._dagit_settings = check.opt_dict_param(dagit_settings, 'dagit_settings')
        self._telemetry_settings = check.opt_dict_param(telemetry_settings, 'telemetry_settings')

        self._ref = check.opt_inst_param(ref, 'ref', InstanceRef)

        self._subscribers = defaultdict(list)

    # ctors

    @staticmethod
    def ephemeral(tempdir=None):
        from dagster.core.storage.event_log import InMemoryEventLogStorage
        from dagster.core.storage.root import LocalArtifactStorage
        from dagster.core.storage.runs import InMemoryRunStorage
        from dagster.core.storage.local_compute_log_manager import NoOpComputeLogManager

        if tempdir is None:
            tempdir = DagsterInstance.temp_storage()

        return DagsterInstance(
            InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(tempdir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            compute_log_manager=NoOpComputeLogManager(compute_logs_directory(tempdir)),
        )

    @staticmethod
    def get(fallback_storage=None):
        # 1. Use $DAGSTER_HOME to determine instance if set.
        if _is_dagster_home_set():
            return DagsterInstance.from_config(_dagster_home())

        # 2. If that is not set use the fallback storage directory if provided.
        # This allows us to have a nice out of the box dagit experience where runs are persisted
        # across restarts in a tempdir that gets cleaned up when the dagit watchdog process exits.
        elif fallback_storage is not None:
            return DagsterInstance.from_config(fallback_storage)

        # 3. If all else fails create an ephemeral in memory instance.
        else:
            return DagsterInstance.ephemeral(fallback_storage)

    @staticmethod
    def local_temp(tempdir=None, overrides=None):
        if tempdir is None:
            tempdir = DagsterInstance.temp_storage()

        return DagsterInstance.from_ref(InstanceRef.from_dir(tempdir, overrides=overrides))

    @staticmethod
    def from_config(config_dir, config_filename=DAGSTER_CONFIG_YAML_FILENAME):
        instance_ref = InstanceRef.from_dir(config_dir, config_filename=config_filename)
        return DagsterInstance.from_ref(instance_ref)

    @staticmethod
    def from_ref(instance_ref):
        check.inst_param(instance_ref, 'instance_ref', InstanceRef)
        return DagsterInstance(
            instance_type=InstanceType.PERSISTENT,
            local_artifact_storage=instance_ref.local_artifact_storage,
            run_storage=instance_ref.run_storage,
            event_storage=instance_ref.event_storage,
            compute_log_manager=instance_ref.compute_log_manager,
            schedule_storage=instance_ref.schedule_storage,
            scheduler=instance_ref.scheduler,
            run_launcher=instance_ref.run_launcher,
            dagit_settings=instance_ref.dagit_settings,
            telemetry_settings=instance_ref.telemetry_settings,
            ref=instance_ref,
        )

    # flags

    @property
    def is_persistent(self):
        return self._instance_type == InstanceType.PERSISTENT

    @property
    def is_ephemeral(self):
        return self._instance_type == InstanceType.EPHEMERAL

    def get_ref(self):
        if self._ref:
            return self._ref

        check.failed('Can not produce an instance reference for {t}'.format(t=self))

    @property
    def root_directory(self):
        return self._local_artifact_storage.base_dir

    @staticmethod
    def temp_storage():
        if DagsterInstance._PROCESS_TEMPDIR is None:
            DagsterInstance._PROCESS_TEMPDIR = seven.TemporaryDirectory()
        return DagsterInstance._PROCESS_TEMPDIR.name

    def _info(self, component):
        prefix = '     '
        # ConfigurableClass may not have inst_data if it's a direct instantiation
        # which happens for ephemeral instances
        if isinstance(component, ConfigurableClass) and component.inst_data:
            return component.inst_data.info_str(prefix)
        if type(component) is dict:
            return prefix + yaml.dump(component, default_flow_style=False).replace(
                '\n', '\n' + prefix
            )
        return '{}{}\n'.format(prefix, component.__class__.__name__)

    def info_str_for_component(self, component_name, component):
        return '{component_name}:\n{component}\n'.format(
            component_name=component_name, component=self._info(component)
        )

    def info_str(self):

        dagit_settings = self._dagit_settings if self._dagit_settings else None
        telemetry_settings = self._telemetry_settings if self._telemetry_settings else None

        return (
            'DagsterInstance components:\n\n'
            '  Local Artifacts Storage:\n{artifact}\n'
            '  Run Storage:\n{run}\n'
            '  Event Log Storage:\n{event}\n'
            '  Compute Log Manager:\n{compute}\n'
            '  Schedule Storage:\n{schedule_storage}\n'
            '  Scheduler:\n{scheduler}\n'
            '  Run Launcher:\n{run_launcher}\n'
            '  Dagit:\n{dagit}\n'
            '  Telemetry:\n{telemetry}\n'
            ''.format(
                artifact=self._info(self._local_artifact_storage),
                run=self._info(self._run_storage),
                event=self._info(self._event_storage),
                compute=self._info(self._compute_log_manager),
                schedule_storage=self._info(self._schedule_storage),
                scheduler=self._info(self._scheduler),
                run_launcher=self._info(self._run_launcher),
                dagit=self._info(dagit_settings),
                telemetry=self._info(telemetry_settings),
            )
        )

    # schedule storage

    @property
    def schedule_storage(self):
        return self._schedule_storage

    # schedule storage

    @property
    def scheduler(self):
        return self._scheduler

    # run launcher

    @property
    def run_launcher(self):
        return self._run_launcher

    # compute logs

    @property
    def compute_log_manager(self):
        return self._compute_log_manager

    @property
    def dagit_settings(self):
        if self._dagit_settings:
            return self._dagit_settings
        return {}

    @property
    def telemetry_enabled(self):
        if self.is_ephemeral:
            return False

        dagster_telemetry_enabled_default = True

        if not self._telemetry_settings:
            return dagster_telemetry_enabled_default

        if 'enabled' in self._telemetry_settings:
            return self._telemetry_settings['enabled']
        else:
            return dagster_telemetry_enabled_default

    def upgrade(self, print_fn=lambda _: None):
        print_fn('Updating run storage...')
        self._run_storage.upgrade()

        print_fn('Updating event storage...')
        self._event_storage.upgrade()

    def dispose(self):
        self._run_storage.dispose()
        self._event_storage.dispose()

    # run storage

    def get_run_by_id(self, run_id):
        return self._run_storage.get_run_by_id(run_id)

    def get_pipeline_snapshot(self, snapshot_id):
        return self._run_storage.get_pipeline_snapshot(snapshot_id)

    def has_pipeline_snapshot(self, snapshot_id):
        return self._run_storage.has_pipeline_snapshot(snapshot_id)

    def get_execution_plan_snapshot(self, snapshot_id):
        return self._run_storage.get_execution_plan_snapshot(snapshot_id)

    def get_run_stats(self, run_id):
        return self._event_storage.get_stats_for_run(run_id)

    def get_run_step_stats(self, run_id):
        return self._event_storage.get_step_stats_for_run(run_id)

    def get_run_tags(self):
        return self._run_storage.get_run_tags()

    def get_run_group(self, run_id):
        return self._run_storage.get_run_group(run_id)

    def create_run_for_pipeline(
        self,
        pipeline_def,
        execution_plan=None,
        run_id=None,
        environment_dict=None,
        mode=None,
        solid_subset=None,
        step_keys_to_execute=None,
        status=None,
        tags=None,
        root_run_id=None,
        parent_run_id=None,
    ):
        from dagster.core.execution.api import create_execution_plan
        from dagster.core.execution.plan.plan import ExecutionPlan
        from dagster.core.snap import snapshot_from_execution_plan

        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.opt_inst_param(execution_plan, 'execution_plan', ExecutionPlan)

        if solid_subset:
            pipeline_def = pipeline_def.build_sub_pipeline(solid_subset=solid_subset)

        if execution_plan is None:
            execution_plan = create_execution_plan(
                pipeline_def,
                environment_dict=environment_dict,
                mode=mode,
                step_keys_to_execute=step_keys_to_execute,
            )

        return self.get_or_create_run(
            pipeline_name=pipeline_def.name,
            run_id=run_id,
            environment_dict=environment_dict,
            mode=check.opt_str_param(mode, 'mode', default=pipeline_def.get_default_mode_name()),
            solid_subset=solid_subset,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            pipeline_snapshot=pipeline_def.get_pipeline_snapshot(),
            execution_plan_snapshot=snapshot_from_execution_plan(
                execution_plan, pipeline_def.get_pipeline_snapshot_id()
            ),
        )

    def _construct_run_with_snapshots(
        self,
        pipeline_name=None,
        run_id=None,
        environment_dict=None,
        mode=None,
        solid_subset=None,
        step_keys_to_execute=None,
        status=None,
        tags=None,
        root_run_id=None,
        parent_run_id=None,
        pipeline_snapshot=None,
        execution_plan_snapshot=None,
    ):

        if tags and IS_AIRFLOW_INGEST_PIPELINE_STR in tags:
            if AIRFLOW_EXECUTION_DATE_STR not in tags:
                tags[AIRFLOW_EXECUTION_DATE_STR] = get_current_datetime_in_utc().isoformat()

        pipeline_run = PipelineRun(
            pipeline_name=pipeline_name,
            run_id=run_id,
            environment_dict=environment_dict,
            mode=mode,
            solid_subset=solid_subset,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
        )

        if pipeline_snapshot is not None:
            from dagster.core.snap import create_pipeline_snapshot_id

            pipeline_snapshot_id = create_pipeline_snapshot_id(pipeline_snapshot)

            if not self._run_storage.has_pipeline_snapshot(pipeline_snapshot_id):
                returned_pipeline_snapshot_id = self._run_storage.add_pipeline_snapshot(
                    pipeline_snapshot
                )

                check.invariant(pipeline_snapshot_id == returned_pipeline_snapshot_id)

            pipeline_run = pipeline_run.with_pipeline_snapshot_id(pipeline_snapshot_id)

        if execution_plan_snapshot is not None:
            from dagster.core.snap import create_execution_plan_snapshot_id

            check.invariant(execution_plan_snapshot.pipeline_snapshot_id == pipeline_snapshot_id)

            execution_plan_snapshot_id = create_execution_plan_snapshot_id(execution_plan_snapshot)

            if not self._run_storage.has_execution_plan_snapshot(execution_plan_snapshot_id):
                returned_execution_plan_snapshot_id = self._run_storage.add_execution_plan_snapshot(
                    execution_plan_snapshot
                )

                check.invariant(execution_plan_snapshot_id == returned_execution_plan_snapshot_id)

            pipeline_run = pipeline_run.with_execution_plan_snapshot_id(execution_plan_snapshot_id)

        return pipeline_run

    def create_run(
        self,
        pipeline_name=None,
        run_id=None,
        environment_dict=None,
        mode=None,
        solid_subset=None,
        step_keys_to_execute=None,
        status=None,
        tags=None,
        root_run_id=None,
        parent_run_id=None,
        pipeline_snapshot=None,
        execution_plan_snapshot=None,
    ):

        # This logic is factored out since it's shared with `get_or_create_run`, which will soon be
        # deleted. When it is deleted, the body of `_construct_run_with_snapshots` can be brought
        # back into this function.
        # https://github.com/dagster-io/dagster/issues/2412
        pipeline_run = self._construct_run_with_snapshots(
            pipeline_name=pipeline_name,
            run_id=run_id,
            environment_dict=environment_dict,
            mode=mode,
            solid_subset=solid_subset,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            pipeline_snapshot=pipeline_snapshot,
            execution_plan_snapshot=execution_plan_snapshot,
        )
        return self._run_storage.add_run(pipeline_run)

    def get_or_create_run(
        self,
        pipeline_name=None,
        run_id=None,
        environment_dict=None,
        mode=None,
        solid_subset=None,
        step_keys_to_execute=None,
        status=None,
        tags=None,
        root_run_id=None,
        parent_run_id=None,
        pipeline_snapshot=None,
        execution_plan_snapshot=None,
    ):
        # The last usage of this method is in dagster-airflow. When the usage is removed, this method
        # should be deleted.
        # https://github.com/dagster-io/dagster/issues/2412

        pipeline_run = self._construct_run_with_snapshots(
            pipeline_name=pipeline_name,
            run_id=run_id,
            environment_dict=environment_dict,
            mode=mode,
            solid_subset=solid_subset,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            pipeline_snapshot=pipeline_snapshot,
            execution_plan_snapshot=execution_plan_snapshot,
        )

        if self.has_run(pipeline_run.run_id):
            candidate_run = self.get_run_by_id(pipeline_run.run_id)

            field_diff = _check_run_equality(pipeline_run, candidate_run)

            if field_diff:
                raise DagsterRunConflict(
                    'Found conflicting existing run with same id {run_id}. Runs differ in:'
                    '\n{field_diff}'.format(
                        run_id=pipeline_run.run_id, field_diff=_format_field_diff(field_diff),
                    ),
                )
            return candidate_run

        return self._run_storage.add_run(pipeline_run)

    def add_run(self, pipeline_run):
        return self._run_storage.add_run(pipeline_run)

    def handle_run_event(self, run_id, event):
        return self._run_storage.handle_run_event(run_id, event)

    def has_run(self, run_id):
        return self._run_storage.has_run(run_id)

    def get_runs(self, filters=None, cursor=None, limit=None):
        return self._run_storage.get_runs(filters, cursor, limit)

    def get_runs_count(self, filters=None):
        return self._run_storage.get_runs_count(filters)

    def wipe(self):
        self._run_storage.wipe()
        self._event_storage.wipe()

    def delete_run(self, run_id):
        self._run_storage.delete_run(run_id)
        self._event_storage.delete_events(run_id)

    # event storage

    def logs_after(self, run_id, cursor):
        return self._event_storage.get_logs_for_run(run_id, cursor=cursor)

    def all_logs(self, run_id):
        return self._event_storage.get_logs_for_run(run_id)

    def watch_event_logs(self, run_id, cursor, cb):
        return self._event_storage.watch(run_id, cursor, cb)

    # event subscriptions

    def get_logger(self):
        logger = logging.Logger('__event_listener')
        logger.addHandler(_EventListenerLogHandler(self))
        logger.setLevel(10)
        return logger

    def handle_new_event(self, event):
        run_id = event.run_id

        self._event_storage.store_event(event)

        if event.is_dagster_event and event.dagster_event.is_pipeline_event:
            self._run_storage.handle_run_event(run_id, event.dagster_event)

        for sub in self._subscribers[run_id]:
            sub(event)

    def add_event_listener(self, run_id, cb):
        self._subscribers[run_id].append(cb)

    def report_engine_event(
        self, message, pipeline_run, engine_event_data=None, cls=None, step_key=None,
    ):
        '''
        Report a EngineEvent that occurred outside of a pipeline execution context.
        '''
        from dagster.core.events import EngineEventData, DagsterEvent, DagsterEventType
        from dagster.core.events.log import DagsterEventRecord

        check.class_param(cls, 'cls')
        check.str_param(message, 'message')
        check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
        engine_event_data = check.opt_inst_param(
            engine_event_data, 'engine_event_data', EngineEventData, EngineEventData([]),
        )

        if cls:
            message = "[{}] {}".format(cls.__name__, message)

        log_level = logging.INFO
        if engine_event_data and engine_event_data.error:
            log_level = logging.ERROR

        dagster_event = DagsterEvent(
            event_type_value=DagsterEventType.ENGINE_EVENT.value,
            pipeline_name=pipeline_run.pipeline_name,
            message=message,
            event_specific_data=engine_event_data,
        )
        event_record = DagsterEventRecord(
            message=message,
            user_message=message,
            level=log_level,
            pipeline_name=pipeline_run.pipeline_name,
            run_id=pipeline_run.run_id,
            error_info=None,
            timestamp=time.time(),
            step_key=step_key,
            dagster_event=dagster_event,
        )

        self.handle_new_event(event_record)
        return dagster_event

    def report_run_failed(self, pipeline_run):
        from dagster.core.events import DagsterEvent, DagsterEventType
        from dagster.core.events.log import DagsterEventRecord

        check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
        message = "This pipeline run has been marked as failed from outside the execution context"

        event_record = DagsterEventRecord(
            message=message,
            user_message=message,
            level=logging.ERROR,
            pipeline_name=pipeline_run.pipeline_name,
            run_id=pipeline_run.run_id,
            error_info=None,
            timestamp=time.time(),
            dagster_event=DagsterEvent(
                event_type_value=DagsterEventType.PIPELINE_FAILURE.value,
                pipeline_name=pipeline_run.pipeline_name,
                message=message,
            ),
        )
        self.handle_new_event(event_record)

    # directories

    def file_manager_directory(self, run_id):
        return self._local_artifact_storage.file_manager_dir(run_id)

    def intermediates_directory(self, run_id):
        return self._local_artifact_storage.intermediates_dir(run_id)

    def schedules_directory(self):
        return self._local_artifact_storage.schedules_dir

    # Run launcher

    def launch_run(self, run_id):
        '''Launch a pipeline run.

        This method delegates to the ``RunLauncher``, if any, configured on the instance, and will
        call its implementation of ``RunLauncher.launch_run()`` to begin the execution of the
        specified run. Runs should be created in the instance (e.g., by calling
        ``DagsterInstance.get_or_create_run()``) *before* this method is called, and
        should be in the ``PipelineRunStatus.NOT_STARTED`` state.

        Args:
            run_id (str): The id of the run the launch.
        '''
        run = self.get_run_by_id(run_id)
        return self._run_launcher.launch_run(self, run)

    # Scheduler

    def start_schedule(self, repository, schedule_name):
        return self._scheduler.start_schedule(self, repository, schedule_name)

    def stop_schedule(self, repository, schedule_name):
        return self._scheduler.stop_schedule(self, repository, schedule_name)

    def end_schedule(self, repository, schedule_name):
        return self._scheduler.end_schedule(self, repository, schedule_name)

    def scheduler_debug_info(self):
        output = ""

        title = "Scheduler Configuration"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info=self.info_str_for_component('Scheduler', self.scheduler),
        )

        title = "Scheduler Info"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title, sep="=" * len(title), info=self.scheduler.debug_info(),
        )

        title = "Scheduler Storage Info"
        schedule_info = self.all_schedules_info()
        schedules = []
        for repository, schedule in schedule_info:
            schedule_info = {
                schedule.name: {
                    "status": schedule.status.value,
                    "cron_schedule": schedule.cron_schedule,
                    "python_path": schedule.python_path,
                    "repository_name": repository,
                    "repository_path": schedule.repository_path,
                }
            }

            schedules.append(yaml.safe_dump(schedule_info, default_flow_style=False))

        output += "{title}\n{sep}\n{info}\n".format(
            title=title, sep="=" * len(title), info="\n".join(schedules),
        )

        return output

    # Schedule Storage

    def create_schedule_tick(self, repository, schedule_tick_data):
        return self._schedule_storage.create_schedule_tick(repository, schedule_tick_data)

    def update_schedule_tick(self, repository, tick):
        return self._schedule_storage.update_schedule_tick(repository, tick)

    def get_schedule_ticks_by_schedule(self, repository, schedule_name):
        return self._schedule_storage.get_schedule_ticks_by_schedule(repository, schedule_name)

    def get_schedule_tick_stats_by_schedule(self, repository, schedule_name):
        return self._schedule_storage.get_schedule_tick_stats_by_schedule(repository, schedule_name)

    def all_schedules_info(self):
        return self._schedule_storage.all_schedules_info()

    def all_schedules(self, repository=None):
        return self._schedule_storage.all_schedules(repository)

    def get_schedule_by_name(self, repository, schedule_name):
        return self._schedule_storage.get_schedule_by_name(repository, schedule_name)

    def add_schedule(self, repository, schedule):
        return self._schedule_storage.add_schedule(repository, schedule)

    def update_schedule(self, repository, schedule):
        return self._schedule_storage.update_schedule(repository, schedule)

    def delete_schedule(self, repository, schedule):
        return self._schedule_storage.delete_schedule(repository, schedule)

    def wipe_all_schedules(self):
        if self._scheduler:
            self._scheduler.wipe(self)

        self._schedule_storage.wipe()

    def log_path_for_schedule(self, repository, schedule_name):
        return self._scheduler.get_log_path(self, repository, schedule_name)



from __future__ import print_function

import os
import random
import string
import sys
import time

import pytest

from dagster import (
    DagsterEventType,
    ExecutionTargetHandle,
    InputDefinition,
    ModeDefinition,
    execute_pipeline,
    pipeline,
    resource,
    solid,
)
from dagster.core.execution.compute_logs import should_disable_io_stream_redirect
from dagster.core.instance import DagsterInstance
from dagster.core.storage.compute_log_manager import ComputeIOType
from dagster.utils import get_multiprocessing_context

HELLO_SOLID = 'HELLO SOLID'
HELLO_RESOURCE = 'HELLO RESOURCE'
SEPARATOR = os.linesep if (os.name == 'nt' and sys.version_info < (3,)) else '\n'


@resource
def resource_a(_):
    print(HELLO_RESOURCE)
    return 'A'


@solid
def spawn(_):
    return 1


@solid(input_defs=[InputDefinition('num', int)], required_resource_keys={'a'})
def spew(_, num):
    print(HELLO_SOLID)
    return num


def define_pipeline():
    @pipeline(mode_defs=[ModeDefinition(resource_defs={'a': resource_a})])
    def spew_pipeline():
        spew(spew(spawn()))

    return spew_pipeline


def normalize_file_content(s):
    return '\n'.join([line for line in s.replace(os.linesep, '\n').split('\n') if line])


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_compute_log_to_disk():
    spew_pipeline = define_pipeline()
    instance = DagsterInstance.local_temp()
    manager = instance.compute_log_manager
    result = execute_pipeline(spew_pipeline, instance=instance)
    assert result.success

    compute_steps = [
        event.step_key
        for event in result.step_event_list
        if event.event_type == DagsterEventType.STEP_START
    ]
    for step_key in compute_steps:
        if step_key.startswith('spawn'):
            continue
        compute_io_path = manager.get_local_path(result.run_id, step_key, ComputeIOType.STDOUT)
        assert os.path.exists(compute_io_path)
        with open(compute_io_path, 'r') as stdout_file:
            assert normalize_file_content(stdout_file.read()) == HELLO_SOLID


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_compute_log_to_disk_multiprocess():
    spew_pipeline = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_pipeline'
    ).build_pipeline_definition()
    instance = DagsterInstance.local_temp()
    manager = instance.compute_log_manager
    result = execute_pipeline(
        spew_pipeline,
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=instance,
    )
    assert result.success

    compute_steps = [
        event.step_key
        for event in result.step_event_list
        if event.event_type == DagsterEventType.STEP_START
    ]
    for step_key in compute_steps:
        if step_key.startswith('spawn'):
            continue
        compute_io_path = manager.get_local_path(result.run_id, step_key, ComputeIOType.STDOUT)
        assert os.path.exists(compute_io_path)
        with open(compute_io_path, 'r') as stdout_file:
            assert normalize_file_content(stdout_file.read()) == HELLO_SOLID


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_compute_log_manager():
    instance = DagsterInstance.local_temp()
    manager = instance.compute_log_manager
    spew_pipeline = define_pipeline()
    result = execute_pipeline(spew_pipeline, instance=instance)
    assert result.success
    compute_steps = [
        event.step_key
        for event in result.step_event_list
        if event.event_type == DagsterEventType.STEP_START
    ]
    assert len(compute_steps) == 3
    step_key = 'spew.compute'
    assert manager.is_watch_completed(result.run_id, step_key)

    stdout = manager.read_logs_file(result.run_id, step_key, ComputeIOType.STDOUT)
    assert normalize_file_content(stdout.data) == HELLO_SOLID

    stderr = manager.read_logs_file(result.run_id, step_key, ComputeIOType.STDERR)
    cleaned_logs = stderr.data.replace('\x1b[34m', '').replace('\x1b[0m', '')
    assert 'dagster - DEBUG - spew_pipeline - ' in cleaned_logs

    bad_logs = manager.read_logs_file('not_a_run_id', step_key, ComputeIOType.STDOUT)
    assert bad_logs.data is None
    assert not manager.is_watch_completed('not_a_run_id', step_key)


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_compute_log_manager_subscriptions():
    instance = DagsterInstance.local_temp()
    spew_pipeline = define_pipeline()
    step_key = 'spew.compute'
    result = execute_pipeline(spew_pipeline, instance=instance)
    stdout_observable = instance.compute_log_manager.observable(
        result.run_id, step_key, ComputeIOType.STDOUT
    )
    stderr_observable = instance.compute_log_manager.observable(
        result.run_id, step_key, ComputeIOType.STDERR
    )
    stdout = []
    stdout_observable.subscribe(stdout.append)
    stderr = []
    stderr_observable.subscribe(stderr.append)
    assert len(stdout) == 1
    assert stdout[0].data.startswith(HELLO_SOLID)
    assert stdout[0].cursor in [12, 13]
    assert len(stderr) == 1
    assert stderr[0].cursor == len(stderr[0].data)
    assert stderr[0].cursor > 400


def gen_solid_name(length):
    return ''.join(random.choice(string.ascii_lowercase) for x in range(length))


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_long_solid_names():
    solid_name = gen_solid_name(300)

    @pipeline(mode_defs=[ModeDefinition(resource_defs={'a': resource_a})])
    def long_pipeline():
        spew.alias(name=solid_name)()

    instance = DagsterInstance.local_temp()
    manager = instance.compute_log_manager

    result = execute_pipeline(
        long_pipeline,
        instance=instance,
        environment_dict={'solids': {solid_name: {'inputs': {'num': 1}}}},
    )
    assert result.success

    compute_steps = [
        event.step_key
        for event in result.step_event_list
        if event.event_type == DagsterEventType.STEP_START
    ]

    assert len(compute_steps) == 1
    step_key = compute_steps[0]
    assert manager.is_watch_completed(result.run_id, step_key)

    stdout = manager.read_logs_file(result.run_id, step_key, ComputeIOType.STDOUT)
    assert normalize_file_content(stdout.data) == HELLO_SOLID


def execute_inner(step_key, pipeline_run, instance_ref):
    instance = DagsterInstance.from_ref(instance_ref)
    inner_step(instance, pipeline_run, step_key)


def inner_step(instance, pipeline_run, step_key):
    with instance.compute_log_manager.watch(pipeline_run, step_key=step_key):
        time.sleep(0.1)
        print(step_key, 'inner 1')
        print(step_key, 'inner 2')
        print(step_key, 'inner 3')
        time.sleep(0.1)


def expected_inner_output(step_key):
    return '\n'.join(
        ["{step_key} inner {num}".format(step_key=step_key, num=i + 1) for i in range(3)]
    )


def expected_outer_prefix():
    return '\n'.join(["outer {num}".format(num=i + 1) for i in range(3)])


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_single():
    instance = DagsterInstance.local_temp()
    pipeline_name = 'foo_pipeline'
    pipeline_run = instance.create_run(pipeline_name=pipeline_name, pipeline_snapshot=None)

    step_keys = ['A', 'B', 'C']

    with instance.compute_log_manager.watch(pipeline_run):
        print('outer 1')
        print('outer 2')
        print('outer 3')

        for step_key in step_keys:
            inner_step(instance, pipeline_run, step_key)

    for step_key in step_keys:
        stdout = instance.compute_log_manager.read_logs_file(
            pipeline_run.run_id, step_key, ComputeIOType.STDOUT
        )
        assert normalize_file_content(stdout.data) == expected_inner_output(step_key)

    full_out = instance.compute_log_manager.read_logs_file(
        pipeline_run.run_id, pipeline_name, ComputeIOType.STDOUT
    )

    assert normalize_file_content(full_out.data).startswith(expected_outer_prefix())


@pytest.mark.skipif(
    should_disable_io_stream_redirect(), reason="compute logs disabled for win / py3.6+"
)
def test_multi():
    instance = DagsterInstance.local_temp()
    pipeline_name = 'foo_pipeline'
    pipeline_run = instance.create_run(pipeline_name=pipeline_name, pipeline_snapshot=None)
    context = get_multiprocessing_context()

    step_keys = ['A', 'B', 'C']

    with instance.compute_log_manager.watch(pipeline_run):
        print('outer 1')
        print('outer 2')
        print('outer 3')

        for step_key in step_keys:
            process = context.Process(
                target=execute_inner, args=(step_key, pipeline_run, instance.get_ref())
            )
            process.start()
            process.join()

    for step_key in step_keys:
        stdout = instance.compute_log_manager.read_logs_file(
            pipeline_run.run_id, step_key, ComputeIOType.STDOUT
        )
        assert normalize_file_content(stdout.data) == expected_inner_output(step_key)

    full_out = instance.compute_log_manager.read_logs_file(
        pipeline_run.run_id, pipeline_name, ComputeIOType.STDOUT
    )

    # The way that the multiprocess compute-logging interacts with pytest (which stubs out the
    # sys.stdout fileno) makes this difficult to test.  The pytest-captured stdout only captures
    # the stdout from the outer process, not also the inner process
    assert normalize_file_content(full_out.data).startswith(expected_outer_prefix())



import graphviz

from dagster import check
from dagster.core.definitions import PipelineDefinition


def build_graphviz_graph(pipeline_def, only_solids):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    graphviz_graph = graphviz.Graph('pipeline_def', directory='/tmp/graphviz')
    for solid in pipeline_def.solids:
        graphviz_graph.node(solid.name)

    graphviz_graph.attr('node', color='grey', shape='box')

    if only_solids:
        for solid in pipeline_def.solids:
            for name in solid.input_dict.keys():
                input_handle = solid.input_handle(name)
                if pipeline_def.dependency_structure.has_deps(input_handle):
                    output_handles = pipeline_def.dependency_structure.get_deps_list(input_handle)
                    for output_handle in output_handles:
                        graphviz_graph.edge(output_handle.solid.name, solid.name)

    else:
        for solid in pipeline_def.solids:
            for name in solid.input_dict.keys():
                scoped_name = solid.name + '.' + name
                graphviz_graph.node(scoped_name)

        for solid in pipeline_def.solids:
            for name in solid.input_dict.keys():
                scoped_name = solid.name + '.' + name
                graphviz_graph.edge(scoped_name, solid.name)
                input_handle = solid.input_handle(name)
                if pipeline_def.dependency_structure.has_deps(input_handle):
                    output_handles = pipeline_def.dependency_structure.get_deps_list(input_handle)
                    for output_handle in output_handles:
                        graphviz_graph.edge(output_handle.solid.name, scoped_name)

    return graphviz_graph



from dagster import ModeDefinition, logger, pipeline, resource
from dagster.core.snap import PipelineSnapshot
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple


def test_mode_snap(snapshot):
    @resource(config={'foo': str})
    def a_resource(_):
        pass

    @resource(description='resource_description')
    def no_config_resource(_):
        pass

    @logger(config={'bar': str})
    def a_logger(_):
        pass

    @logger(description='logger_description')
    def no_config_logger(_):
        pass

    @pipeline(
        mode_defs=[
            ModeDefinition(
                name='a_mode',
                description='a_desc',
                resource_defs={
                    'some_resource': a_resource,
                    'no_config_resource': no_config_resource,
                },
                logger_defs={'some_logger': a_logger, 'no_config_logger': no_config_logger},
            )
        ]
    )
    def a_pipeline():
        pass

    pipeline_snapshot = PipelineSnapshot.from_pipeline_def(a_pipeline)
    assert len(pipeline_snapshot.mode_def_snaps) == 1
    mode_def_snap = pipeline_snapshot.mode_def_snaps[0]

    snapshot.assert_match(serialize_dagster_namedtuple(mode_def_snap))

    assert mode_def_snap == deserialize_json_to_dagster_namedtuple(
        serialize_dagster_namedtuple(mode_def_snap)
    )



import os

from dagster import IntSource, StringSource
from dagster.config.validate import process_config
from dagster.core.test_utils import environ


def test_string_source():
    assert process_config(StringSource, 'foo').success
    assert not process_config(StringSource, 1).success

    assert not process_config(StringSource, {'env': 1}).success

    assert 'DAGSTER_TEST_ENV_VAR' not in os.environ
    assert not process_config(StringSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).success

    assert (
        'You have attempted to fetch the environment variable "DAGSTER_TEST_ENV_VAR" '
        'which is not set. In order for this execution to succeed it must be set in '
        'this environment.'
    ) in process_config(StringSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).errors[0].message

    with environ({'DAGSTER_TEST_ENV_VAR': 'baz'}):
        assert process_config(StringSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).success
        assert process_config(StringSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).value == 'baz'


def test_int_source():
    assert process_config(IntSource, 1).success
    assert not process_config(IntSource, 'foo').success

    assert not process_config(IntSource, {'env': 1}).success

    assert 'DAGSTER_TEST_ENV_VAR' not in os.environ
    assert not process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).success

    assert (
        'You have attempted to fetch the environment variable "DAGSTER_TEST_ENV_VAR" '
        'which is not set. In order for this execution to succeed it must be set in '
        'this environment.'
    ) in process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).errors[0].message

    with environ({'DAGSTER_TEST_ENV_VAR': '4'}):
        assert process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).success
        assert process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).value == 4

    with environ({'DAGSTER_TEST_ENV_VAR': 'four'}):
        assert not process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).success
        assert (
            'Value "four" stored in env variable "DAGSTER_TEST_ENV_VAR" cannot '
            'be coerced into an int.'
        ) in process_config(IntSource, {'env': 'DAGSTER_TEST_ENV_VAR'}).errors[0].message



import pytest

from dagster import (
    InputDefinition,
    ModeDefinition,
    OutputDefinition,
    String,
    execute_pipeline,
    input_hydration_config,
    output_materialization_config,
    pipeline,
    resource,
    solid,
    usable_as_dagster_type,
)
from dagster.core.storage.type_storage import TypeStoragePlugin
from dagster.core.types.dagster_type import create_any_type


class UserError(Exception):
    def __init__(self):
        super(UserError, self).__init__('The user has errored')


def test_user_error_boundary_solid_compute():
    @solid
    def throws_user_error(_):
        raise UserError()

    @pipeline
    def pipeline_def():
        throws_user_error()

    pipeline_result = execute_pipeline(pipeline_def, raise_on_error=False)
    assert not pipeline_result.success


def test_user_error_boundary_input_hydration():
    @input_hydration_config(String)
    def InputHydration(context, hello):
        raise UserError()

    @usable_as_dagster_type(input_hydration_config=InputHydration)
    class CustomType(str):
        pass

    @solid(input_defs=[InputDefinition('custom_type', CustomType)])
    def input_hydration_solid(context, custom_type):
        context.log.info(custom_type)

    @pipeline
    def input_hydration_pipeline():
        input_hydration_solid()

    pipeline_result = execute_pipeline(
        input_hydration_pipeline,
        {'solids': {'input_hydration_solid': {'inputs': {'custom_type': 'hello'}}}},
        raise_on_error=False,
    )
    assert not pipeline_result.success


def test_user_error_boundary_output_materialization():
    @output_materialization_config(String)
    def materialize(context, *_args, **_kwargs):
        raise UserError()

    CustomDagsterType = create_any_type(
        name='CustomType', output_materialization_config=materialize
    )

    @solid(output_defs=[OutputDefinition(CustomDagsterType)])
    def output_solid(_context):
        return 'hello'

    @pipeline
    def output_materialization_pipeline():
        output_solid()

    pipeline_result = execute_pipeline(
        output_materialization_pipeline,
        {'solids': {'output_solid': {'outputs': [{'result': 'hello'}]}}},
        raise_on_error=False,
    )
    assert not pipeline_result.success


def test_user_error_boundary_resource_init():
    @resource
    def resource_a(_):
        raise UserError()

    @solid(required_resource_keys={'a'})
    def resource_solid(_context):
        return 'hello'

    @pipeline(mode_defs=[ModeDefinition(resource_defs={'a': resource_a})])
    def resource_pipeline():
        resource_solid()

    pipeline_result = execute_pipeline(resource_pipeline, raise_on_error=False)
    assert not pipeline_result.success


@pytest.mark.skip('not implemented')
def test_user_error_boundary_storage_plugin():
    class CustomStoragePlugin(TypeStoragePlugin):  # pylint: disable=no-init
        @classmethod
        def compatible_with_storage_def(cls, system_storage_def):
            return True

        @classmethod
        def set_object(cls, intermediate_store, obj, context, dagster_type, paths):
            raise UserError()

        @classmethod
        def get_object(cls, intermediate_store, context, dagster_type, paths):
            raise UserError()

    CustomDagsterType = create_any_type(name='CustomType', auto_plugins=[CustomStoragePlugin])

    @solid(output_defs=[OutputDefinition(CustomDagsterType)])
    def output_solid(_context):
        return 'hello'

    @pipeline
    def plugin_pipeline():
        output_solid()

    pipeline_result = execute_pipeline(
        plugin_pipeline, {'storage': {'filesystem': {}}}, raise_on_error=False
    )
    assert not pipeline_result.success



from dagster import DagsterInstance, ModeDefinition, PipelineDefinition, resource, solid
from dagster.core.definitions.executable import InMemoryExecutablePipeline
from dagster.core.events.log import EventRecord, construct_event_logger
from dagster.core.execution.api import (
    create_execution_plan,
    execute_pipeline_iterator,
    execute_plan_iterator,
    execute_run_iterator,
)


@resource
def resource_a(context):
    context.log.info('CALLING A')
    yield 'A'
    context.log.info('CLEANING A')
    yield  # add the second yield here to test teardown generator exit handling


@resource
def resource_b(context):
    context.log.info('CALLING B')
    yield 'B'
    context.log.info('CLEANING B')
    yield  # add the second yield here to test teardown generator exit handling


@solid(required_resource_keys={'a', 'b'})
def resource_solid(_):
    return 'A'


def test_execute_pipeline_iterator():
    records = []

    def event_callback(record):
        assert isinstance(record, EventRecord)
        records.append(record)

    pipeline = PipelineDefinition(
        name='basic_resource_pipeline',
        solid_defs=[resource_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={'a': resource_a, 'b': resource_b},
                logger_defs={'callback': construct_event_logger(event_callback)},
            )
        ],
    )
    iterator = execute_pipeline_iterator(
        pipeline,
        environment_dict={'loggers': {'callback': {}}},
        instance=DagsterInstance.local_temp(),
    )

    event_type = None
    while event_type != 'STEP_START':
        event = next(iterator)
        event_type = event.event_type_value

    iterator.close()
    events = [record.dagster_event for record in records if record.is_dagster_event]
    messages = [record.user_message for record in records if not record.is_dagster_event]
    assert len([event for event in events if event.is_pipeline_failure]) > 0
    assert len([message for message in messages if message == 'CLEANING A']) > 0
    assert len([message for message in messages if message == 'CLEANING B']) > 0


def test_execute_run_iterator():
    records = []

    def event_callback(record):
        assert isinstance(record, EventRecord)
        records.append(record)

    instance = DagsterInstance.local_temp()

    pipeline_def = PipelineDefinition(
        name='basic_resource_pipeline',
        solid_defs=[resource_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={'a': resource_a, 'b': resource_b},
                logger_defs={'callback': construct_event_logger(event_callback)},
            )
        ],
    )
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def, environment_dict={'loggers': {'callback': {}}}, mode='default',
    )

    iterator = execute_run_iterator(
        InMemoryExecutablePipeline(pipeline_def), pipeline_run, instance=instance
    )

    event_type = None
    while event_type != 'STEP_START':
        event = next(iterator)
        event_type = event.event_type_value

    iterator.close()
    events = [record.dagster_event for record in records if record.is_dagster_event]
    messages = [record.user_message for record in records if not record.is_dagster_event]
    assert len([event for event in events if event.is_pipeline_failure]) > 0
    assert len([message for message in messages if message == 'CLEANING A']) > 0
    assert len([message for message in messages if message == 'CLEANING B']) > 0


def test_execute_plan_iterator():
    records = []

    def event_callback(record):
        assert isinstance(record, EventRecord)
        records.append(record)

    instance = DagsterInstance.local_temp()

    pipeline = PipelineDefinition(
        name='basic_resource_pipeline',
        solid_defs=[resource_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={'a': resource_a, 'b': resource_b},
                logger_defs={'callback': construct_event_logger(event_callback)},
            )
        ],
    )
    environment_dict = {'loggers': {'callback': {}}}

    execution_plan = create_execution_plan(pipeline, environment_dict=environment_dict)
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline,
        environment_dict={'loggers': {'callback': {}}},
        execution_plan=execution_plan,
    )

    iterator = execute_plan_iterator(
        execution_plan, pipeline_run, instance, environment_dict=environment_dict
    )

    event_type = None
    while event_type != 'STEP_START':
        event = next(iterator)
        event_type = event.event_type_value

    iterator.close()
    messages = [record.user_message for record in records if not record.is_dagster_event]
    assert len([message for message in messages if message == 'CLEANING A']) > 0
    assert len([message for message in messages if message == 'CLEANING B']) > 0



import pytest

from dagster import (
    DagsterEventType,
    DagsterResourceFunctionError,
    ExecutionTargetHandle,
    Field,
    Int,
    ModeDefinition,
    PipelineDefinition,
    ResourceDefinition,
    String,
    execute_pipeline,
    execute_pipeline_iterator,
    resource,
    solid,
)
from dagster.core.events.log import EventRecord, LogMessageRecord, construct_event_logger
from dagster.core.execution.api import create_execution_plan, execute_plan
from dagster.core.execution.config import RunConfig
from dagster.core.instance import DagsterInstance


def define_string_resource():
    return ResourceDefinition(
        config=String, resource_fn=lambda init_context: init_context.resource_config
    )


def test_basic_resource():
    called = {}

    @solid(required_resource_keys={'a_string'})
    def a_solid(context):
        called['yup'] = True
        assert context.resources.a_string == 'foo'

    pipeline_def = PipelineDefinition(
        name='with_a_resource',
        solid_defs=[a_solid],
        mode_defs=[ModeDefinition(resource_defs={'a_string': define_string_resource()})],
    )

    result = execute_pipeline(pipeline_def, {'resources': {'a_string': {'config': 'foo'}}})

    assert result.success
    assert called['yup']


def test_yield_resource():
    called = {}

    @solid(required_resource_keys={'a_string'})
    def a_solid(context):
        called['yup'] = True
        assert context.resources.a_string == 'foo'

    def _do_resource(init_context):
        yield init_context.resource_config

    yield_string_resource = ResourceDefinition(config=String, resource_fn=_do_resource)

    pipeline_def = PipelineDefinition(
        name='with_a_yield_resource',
        solid_defs=[a_solid],
        mode_defs=[ModeDefinition(resource_defs={'a_string': yield_string_resource})],
    )

    result = execute_pipeline(pipeline_def, {'resources': {'a_string': {'config': 'foo'}}})

    assert result.success
    assert called['yup']


def test_yield_multiple_resources():
    called = {}

    saw = []

    @solid(required_resource_keys={'string_one', 'string_two'})
    def a_solid(context):
        called['yup'] = True
        assert context.resources.string_one == 'foo'
        assert context.resources.string_two == 'bar'

    def _do_resource(init_context):
        saw.append('before yield ' + init_context.resource_config)
        yield init_context.resource_config
        saw.append('after yield ' + init_context.resource_config)

    yield_string_resource = ResourceDefinition(config=String, resource_fn=_do_resource)

    pipeline_def = PipelineDefinition(
        name='with_yield_resources',
        solid_defs=[a_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'string_one': yield_string_resource,
                    'string_two': yield_string_resource,
                }
            )
        ],
    )

    result = execute_pipeline(
        pipeline_def,
        {'resources': {'string_one': {'config': 'foo'}, 'string_two': {'config': 'bar'}}},
    )

    assert result.success
    assert called['yup']
    assert len(saw) == 4

    assert 'before yield' in saw[0]
    assert 'before yield' in saw[1]
    assert 'after yield' in saw[2]
    assert 'after yield' in saw[3]


def test_resource_decorator():
    called = {}

    saw = []

    @solid(required_resource_keys={'string_one', 'string_two'})
    def a_solid(context):
        called['yup'] = True
        assert context.resources.string_one == 'foo'
        assert context.resources.string_two == 'bar'

    # API red alert. One has to wrap a type in Field because it is callable
    @resource(config=Field(String))
    def yielding_string_resource(init_context):
        saw.append('before yield ' + init_context.resource_config)
        yield init_context.resource_config
        saw.append('after yield ' + init_context.resource_config)

    pipeline_def = PipelineDefinition(
        name='with_yield_resources',
        solid_defs=[a_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'string_one': yielding_string_resource,
                    'string_two': yielding_string_resource,
                }
            )
        ],
    )

    result = execute_pipeline(
        pipeline_def,
        {'resources': {'string_one': {'config': 'foo'}, 'string_two': {'config': 'bar'}}},
    )

    assert result.success
    assert called['yup']
    assert len(saw) == 4

    assert 'before yield' in saw[0]
    assert 'before yield' in saw[1]
    assert 'after yield' in saw[2]
    assert 'after yield' in saw[3]


def test_mixed_multiple_resources():
    called = {}

    saw = []

    @solid(required_resource_keys={'returned_string', 'yielded_string'})
    def a_solid(context):
        called['yup'] = True
        assert context.resources.returned_string == 'foo'
        assert context.resources.yielded_string == 'bar'

    def _do_yield_resource(init_context):
        saw.append('before yield ' + init_context.resource_config)
        yield init_context.resource_config
        saw.append('after yield ' + init_context.resource_config)

    yield_string_resource = ResourceDefinition(config=String, resource_fn=_do_yield_resource)

    def _do_return_resource(init_context):
        saw.append('before return ' + init_context.resource_config)
        return init_context.resource_config

    return_string_resource = ResourceDefinition(config=String, resource_fn=_do_return_resource)

    pipeline_def = PipelineDefinition(
        name='with_a_yield_resource',
        solid_defs=[a_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'yielded_string': yield_string_resource,
                    'returned_string': return_string_resource,
                }
            )
        ],
    )

    result = execute_pipeline(
        pipeline_def,
        {'resources': {'returned_string': {'config': 'foo'}, 'yielded_string': {'config': 'bar'}}},
    )

    assert result.success
    assert called['yup']
    # could be processed in any order in python 2
    assert 'before yield bar' in saw[0] or 'before return foo' in saw[0]
    assert 'before yield bar' in saw[1] or 'before return foo' in saw[1]
    assert 'after yield bar' in saw[2]


def test_none_resource():
    called = {}

    @solid(required_resource_keys={'test_null'})
    def solid_test_null(context):
        assert context.resources.test_null is None
        called['yup'] = True

    pipeline = PipelineDefinition(
        name='test_none_resource',
        solid_defs=[solid_test_null],
        mode_defs=[ModeDefinition(resource_defs={'test_null': ResourceDefinition.none_resource()})],
    )

    result = execute_pipeline(pipeline)

    assert result.success
    assert called['yup']


def test_string_resource():
    called = {}

    @solid(required_resource_keys={'test_string'})
    def solid_test_string(context):
        assert context.resources.test_string == 'foo'
        called['yup'] = True

    pipeline = PipelineDefinition(
        name='test_string_resource',
        solid_defs=[solid_test_string],
        mode_defs=[
            ModeDefinition(resource_defs={'test_string': ResourceDefinition.string_resource()})
        ],
    )

    result = execute_pipeline(pipeline, {'resources': {'test_string': {'config': 'foo'}}})

    assert result.success
    assert called['yup']


def test_no_config_resource_pass_none():
    called = {}

    @resource(None)
    def return_thing(_init_context):
        called['resource'] = True
        return 'thing'

    @solid(required_resource_keys={'return_thing'})
    def check_thing(context):
        called['solid'] = True
        assert context.resources.return_thing == 'thing'

    pipeline = PipelineDefinition(
        name='test_no_config_resource',
        solid_defs=[check_thing],
        mode_defs=[ModeDefinition(resource_defs={'return_thing': return_thing})],
    )

    execute_pipeline(pipeline)

    assert called['resource']
    assert called['solid']


def test_no_config_resource_no_arg():
    called = {}

    @resource()
    def return_thing(_init_context):
        called['resource'] = True
        return 'thing'

    @solid(required_resource_keys={'return_thing'})
    def check_thing(context):
        called['solid'] = True
        assert context.resources.return_thing == 'thing'

    pipeline = PipelineDefinition(
        name='test_no_config_resource',
        solid_defs=[check_thing],
        mode_defs=[ModeDefinition(resource_defs={'return_thing': return_thing})],
    )

    execute_pipeline(pipeline)

    assert called['resource']
    assert called['solid']


def test_no_config_resource_bare_no_arg():
    called = {}

    @resource
    def return_thing(_init_context):
        called['resource'] = True
        return 'thing'

    @solid(required_resource_keys={'return_thing'})
    def check_thing(context):
        called['solid'] = True
        assert context.resources.return_thing == 'thing'

    pipeline = PipelineDefinition(
        name='test_no_config_resource',
        solid_defs=[check_thing],
        mode_defs=[ModeDefinition(resource_defs={'return_thing': return_thing})],
    )

    execute_pipeline(pipeline)

    assert called['resource']
    assert called['solid']


def test_no_config_resource_definition():
    called = {}

    def _return_thing_resource_fn(_init_context):
        called['resource'] = True
        return 'thing'

    @solid(required_resource_keys={'return_thing'})
    def check_thing(context):
        called['solid'] = True
        assert context.resources.return_thing == 'thing'

    pipeline = PipelineDefinition(
        name='test_no_config_resource',
        solid_defs=[check_thing],
        mode_defs=[
            ModeDefinition(
                resource_defs={'return_thing': ResourceDefinition(_return_thing_resource_fn)}
            )
        ],
    )

    execute_pipeline(pipeline)

    assert called['resource']
    assert called['solid']


def test_resource_cleanup():
    called = {}

    def _cleanup_resource_fn(_init_context):
        called['creation'] = True
        yield True
        called['cleanup'] = True

    @solid(required_resource_keys={'resource_with_cleanup'})
    def check_resource_created(context):
        called['solid'] = True
        assert context.resources.resource_with_cleanup is True

    pipeline = PipelineDefinition(
        name='test_resource_cleanup',
        solid_defs=[check_resource_created],
        mode_defs=[
            ModeDefinition(
                resource_defs={'resource_with_cleanup': ResourceDefinition(_cleanup_resource_fn)}
            )
        ],
    )

    execute_pipeline(pipeline)

    assert called['creation'] is True
    assert called['solid'] is True
    assert called['cleanup'] is True


def test_stacked_resource_cleanup():
    called = []

    def _cleanup_resource_fn_1(_init_context):
        called.append('creation_1')
        yield True
        called.append('cleanup_1')

    def _cleanup_resource_fn_2(_init_context):
        called.append('creation_2')
        yield True
        called.append('cleanup_2')

    @solid(required_resource_keys={'resource_with_cleanup_1', 'resource_with_cleanup_2'})
    def check_resource_created(context):
        called.append('solid')
        assert context.resources.resource_with_cleanup_1 is True
        assert context.resources.resource_with_cleanup_2 is True

    pipeline = PipelineDefinition(
        name='test_resource_cleanup',
        solid_defs=[check_resource_created],
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'resource_with_cleanup_1': ResourceDefinition(_cleanup_resource_fn_1),
                    'resource_with_cleanup_2': ResourceDefinition(_cleanup_resource_fn_2),
                }
            )
        ],
    )

    execute_pipeline(pipeline)

    assert called == ['creation_1', 'creation_2', 'solid', 'cleanup_2', 'cleanup_1']


def test_resource_init_failure():
    @resource
    def failing_resource(_init_context):
        raise Exception('Uh oh')

    @solid(required_resource_keys={'failing_resource'})
    def failing_resource_solid(_context):
        pass

    pipeline = PipelineDefinition(
        name='test_resource_init_failure',
        solid_defs=[failing_resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'failing_resource': failing_resource})],
    )

    res = execute_pipeline(pipeline, raise_on_error=False)

    event_types = [event.event_type_value for event in res.event_list]
    assert DagsterEventType.PIPELINE_INIT_FAILURE.value in event_types

    instance = DagsterInstance.ephemeral()
    execution_plan = create_execution_plan(pipeline)
    pipeline_run = instance.create_run_for_pipeline(pipeline, execution_plan=execution_plan)

    step_events = execute_plan(execution_plan, pipeline_run=pipeline_run, instance=instance)

    event_types = [event.event_type_value for event in step_events]
    assert DagsterEventType.PIPELINE_INIT_FAILURE.value in event_types

    # Test the pipeline init failure event fires even if we are raising errors
    events = []
    try:
        for event in execute_pipeline_iterator(pipeline):
            events.append(event)
    except DagsterResourceFunctionError:
        pass

    event_types = [event.event_type_value for event in events]
    assert DagsterEventType.PIPELINE_INIT_FAILURE.value in event_types


def test_dagster_type_resource_decorator_config():
    @resource(Int)
    def dagster_type_resource_config(_):
        raise Exception('not called')

    assert dagster_type_resource_config.config_field.config_type.given_name == 'Int'

    @resource(int)
    def python_type_resource_config(_):
        raise Exception('not called')

    assert python_type_resource_config.config_field.config_type.given_name == 'Int'


def test_resource_init_failure_with_teardown():
    called = []
    cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @resource
    def resource_b(_):
        try:
            called.append('B')
            raise Exception('uh oh')
            yield 'B'  # pylint: disable=unreachable
        finally:
            cleaned.append('B')

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        pass

    pipeline = PipelineDefinition(
        name='test_resource_init_failure_with_cleanup',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )

    res = execute_pipeline(pipeline, raise_on_error=False)
    event_types = [event.event_type_value for event in res.event_list]
    assert DagsterEventType.PIPELINE_INIT_FAILURE.value in event_types

    assert called == ['A', 'B']
    assert cleaned == ['B', 'A']

    called = []
    cleaned = []

    events = []
    try:
        for event in execute_pipeline_iterator(pipeline):
            events.append(event)
    except DagsterResourceFunctionError:
        pass

    event_types = [event.event_type_value for event in events]
    assert DagsterEventType.PIPELINE_INIT_FAILURE.value in event_types
    assert called == ['A', 'B']
    assert cleaned == ['B', 'A']


def test_solid_failure_resource_teardown():
    called = []
    cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @resource
    def resource_b(_):
        try:
            called.append('B')
            yield 'B'
        finally:
            cleaned.append('B')

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        raise Exception('uh oh')

    pipeline = PipelineDefinition(
        name='test_solid_failure_resource_teardown',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )

    res = execute_pipeline(pipeline, raise_on_error=False)
    assert res.event_list[-1].event_type_value == 'PIPELINE_FAILURE'
    assert called == ['A', 'B']
    assert cleaned == ['B', 'A']

    called = []
    cleaned = []

    events = []
    try:
        for event in execute_pipeline_iterator(pipeline):
            events.append(event)
    except DagsterResourceFunctionError:
        pass

    assert len(events) > 1
    assert events[-1].event_type_value == 'PIPELINE_FAILURE'
    assert called == ['A', 'B']
    assert cleaned == ['B', 'A']


def test_solid_failure_resource_teardown_raise():
    ''' test that teardown is invoked in resources for tests that raise_on_error '''
    called = []
    cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @resource
    def resource_b(_):
        try:
            called.append('B')
            yield 'B'
        finally:
            cleaned.append('B')

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        raise Exception('uh oh')

    pipeline = PipelineDefinition(
        name='test_solid_failure_resource_teardown',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )

    with pytest.raises(Exception) as exc_info:
        execute_pipeline(pipeline)

    assert called == ['A', 'B']
    assert cleaned == ['B', 'A']

    called = []
    cleaned = []


def test_resource_teardown_failure():
    called = []
    cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @resource
    def resource_b(_):
        try:
            called.append('B')
            yield 'B'
        finally:
            raise Exception('uh oh')
            cleaned.append('B')  # pylint: disable=unreachable

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        pass

    pipeline = PipelineDefinition(
        name='test_resource_teardown_failure',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )

    result = execute_pipeline(pipeline, raise_on_error=False)
    assert result.success
    error_events = [
        event
        for event in result.event_list
        if event.is_engine_event and event.event_specific_data.error
    ]
    assert len(error_events) == 1
    assert called == ['A', 'B']
    assert cleaned == ['A']

    called = []
    cleaned = []
    events = []
    try:
        for event in execute_pipeline_iterator(pipeline):
            events.append(event)
    except DagsterResourceFunctionError:
        pass

    assert called == ['A', 'B']
    assert cleaned == ['A']


def define_resource_teardown_failure_pipeline():
    @resource
    def resource_a(_):
        try:
            yield 'A'
        finally:
            pass

    @resource
    def resource_b(_):
        try:
            yield 'B'
        finally:
            raise Exception('uh oh')

    @solid(required_resource_keys={'a', 'b'})
    def resource_solid(_):
        pass

    return PipelineDefinition(
        name='resource_teardown_failure',
        solid_defs=[resource_solid],
        mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})],
    )


def test_multiprocessing_resource_teardown_failure():
    pipeline = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_resource_teardown_failure_pipeline'
    ).build_pipeline_definition()
    result = execute_pipeline(
        pipeline,
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
        raise_on_error=False,
    )
    assert result.success
    error_events = [
        event
        for event in result.event_list
        if event.is_engine_event and event.event_specific_data.error
    ]
    assert len(error_events) > 1


def test_single_step_resource_event_logs():
    # Test to attribute logs for single-step plans which are often the representation of
    # sub-plans in a multiprocessing execution environment. Most likely will need to be rewritten
    # with the refactor detailed in https://github.com/dagster-io/dagster/issues/2239
    USER_SOLID_MESSAGE = 'I AM A SOLID'
    USER_RESOURCE_MESSAGE = 'I AM A RESOURCE'
    events = []

    def event_callback(record):
        assert isinstance(record, EventRecord)
        events.append(record)

    @solid(required_resource_keys={'a'})
    def resource_solid(context):
        context.log.info(USER_SOLID_MESSAGE)

    @resource
    def resource_a(context):
        context.log.info(USER_RESOURCE_MESSAGE)
        return 'A'

    pipeline = PipelineDefinition(
        name='resource_logging_pipeline',
        solid_defs=[resource_solid],
        mode_defs=[
            ModeDefinition(
                resource_defs={'a': resource_a},
                logger_defs={'callback': construct_event_logger(event_callback)},
            )
        ],
    )

    result = execute_pipeline(
        pipeline,
        environment_dict={'loggers': {'callback': {}},},
        instance=DagsterInstance.local_temp(),
        run_config=RunConfig(step_keys_to_execute=['resource_solid.compute']),
    )
    assert result.success
    log_messages = [event for event in events if isinstance(event, LogMessageRecord)]
    assert len(log_messages) == 2

    resource_log_message = next(
        iter([message for message in log_messages if message.user_message == USER_RESOURCE_MESSAGE])
    )
    assert resource_log_message.step_key == 'resource_solid.compute'



from collections import defaultdict

import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    PipelineDefinition,
    RepositoryDefinition,
    SolidDefinition,
    lambda_solid,
)


def create_single_node_pipeline(name, called):
    called[name] = called[name] + 1
    return PipelineDefinition(
        name=name,
        solid_defs=[
            SolidDefinition(
                name=name + '_solid',
                input_defs=[],
                output_defs=[],
                compute_fn=lambda *_args, **_kwargs: None,
            )
        ],
    )


def test_repo_lazy_definition():
    called = defaultdict(int)
    repo = RepositoryDefinition(
        name='some_repo',
        pipeline_dict={
            'foo': lambda: create_single_node_pipeline('foo', called),
            'bar': lambda: create_single_node_pipeline('bar', called),
        },
    )

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' not in called

    bar_pipeline = repo.get_pipeline('bar')
    assert isinstance(bar_pipeline, PipelineDefinition)
    assert bar_pipeline.name == 'bar'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' in called
    assert called['bar'] == 1

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1

    pipelines = repo.get_all_pipelines()

    assert set(['foo', 'bar']) == {pipeline.name for pipeline in pipelines}

    assert repo.solid_def_named('foo_solid').name == 'foo_solid'
    assert repo.solid_def_named('bar_solid').name == 'bar_solid'


def test_dupe_solid_repo_definition():
    @lambda_solid(name='same')
    def noop():
        pass

    @lambda_solid(name='same')
    def noop2():
        pass

    repo = RepositoryDefinition(
        'error_repo',
        pipeline_dict={
            'first': lambda: PipelineDefinition(name='first', solid_defs=[noop]),
            'second': lambda: PipelineDefinition(name='second', solid_defs=[noop2]),
        },
    )

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        repo.get_all_pipelines()

    assert str(exc_info.value) == (
        'You have defined two solid definitions named "same" in repository '
        '"error_repo". Solid definition names must be unique within a '
        'repository. The solid definition has been defined '
        'in pipeline "first" and it has been defined again '
        'in pipeline "second."'
    )


def test_non_lazy_pipeline_dict():
    called = defaultdict(int)
    repo = RepositoryDefinition(
        name='some_repo',
        pipeline_defs=[
            create_single_node_pipeline('foo', called),
            create_single_node_pipeline('bar', called),
        ],
    )

    assert repo.get_pipeline('foo').name == 'foo'
    assert repo.get_pipeline('bar').name == 'bar'


def test_conflict():
    called = defaultdict(int)
    with pytest.raises(Exception, match='Duplicate pipelines named foo'):
        RepositoryDefinition(
            name='some_repo',
            pipeline_defs=[create_single_node_pipeline('foo', called)],
            pipeline_dict={'foo': lambda: create_single_node_pipeline('foo', called)},
        )


def test_key_mismatch():
    called = defaultdict(int)
    repo = RepositoryDefinition(
        name='some_repo', pipeline_dict={'foo': lambda: create_single_node_pipeline('bar', called)}
    )
    with pytest.raises(Exception, match='Name does not match'):
        repo.get_pipeline('foo')
